// $Id: AIEventHandler.C,v 1.19 2007/05/27 22:40:37 mbrzozowski Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

// A sample client for ORTS that shows how to access client data

// todo/questions:
//
// game scripts: 
// - what to do with workers during build time
// - similarly, how to define mining?
// - how is attack range defined (distance between centers/hulls)?
//
// client:
// - how to access global player data?
// - how to compute the set of visible tiles?
// - how to compute tiles visible from a unit?
// - how to decide what can be attacked by a specific unit?

// FORS/FORALL/ERR macros are defined in orts3/libs/kernel/src/Global.H
#include <assert.h>
#include <map>
#include <bitset>


#include "GameStateModule.H"
#include "GfxModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameTile.H"
#include "ServerObjData.H"
#include "Options.H"

#include "AIGUI.H"
#include "AIObject.H"
#include "AIEventHandler.H"
#include "AIState.H"
#include "Graph.H"
#include "Coordinator.H"
#include "Obstacle.H"
#include "Mineral.H"
#include "Configuration.H"
#include "Path.H"

using namespace std;
using namespace bzAI;
// tile types from core.bp
// fixme: game script reflection would help here
// in the competition we won't use water tiles
// TILE_CLIFF: non-traversable by ground units

enum { TILE_UNKNOWN=0, TILE_WATER=1, TILE_GROUND=2, TILE_CLIFF=3 };

AIEventHandler* AIEventHandler::handler = NULL;


AIEventHandler::AIEventHandler(AIState* _state, sint4 seed) :state(_state), coordinator(NULL), graph(NULL){
	assert( !handler);
	// cannot initialize graph and coordinator - it has to be done in the
	// first compute_actions call
	rand.srand(seed);
	collector = new AIObjectCollector();
	AIEventHandler::handler = this;

	int gnum;
	Options::get("-game_type", gnum);
	switch (gnum){
		case 1: game_type=GAME_1;config = new Configuration("apps/aiclient1/src/aiclient1g1.cfg"); break;
		case 2: game_type=GAME_2;config = new Configuration("apps/aiclient1/src/aiclient1g2.cfg"); break;
		case 3: game_type=GAME_3; break;
		case 4: game_type=GAME_4;config = new Configuration("apps/aiclient1/src/aiclient1g4.cfg"); break;
		default: ERR("Game type unrecognizable");


	}
	cout << "Playing game "<<HANDLER->game_type<<endl;
	current_frame=0;
}

AIEventHandler::~AIEventHandler(){
	delete graph;
	delete coordinator;
	delete collector;
	delete config;
	AIEventHandler::handler = NULL;
}


bool AIEventHandler::handle_event(const Event& e)
{
	if (e.get_who() == GameStateModule::FROM) {

		//cout << "Event: " << e.who() << " " << e.what() << endl;

		if (e.get_what() == GameStateModule::STOP_MSG) {
			state->quit = true; return true;
		}

		if (e.get_what() == GameStateModule::READ_ERROR_MSG) {
			state->quit = state->error = true; return true;
		}

		if (e.get_what() == GameStateModule::VIEW_MSG) {


			compute_actions();
			state->gsm->send_actions();


			Game* game = &state->gsm->get_game();	

			if (state->gfxm)
				state->gfxm->process_changes();

			// gfx and frame statistics
			sint4 vf = game->get_view_frame();
			sint4 af = game->get_action_frame();
			sint4 sa = game->get_skipped_actions();
			sint4 ma = game->get_merged_actions();
                        if (vf % 10 == 0){

                          cout << "frame " << vf;
                          if (af < 0) cout << " [no action]";
                          else if (vf != af) cout << " [behind by " << (vf - af) << " frame(s)]";
                          if (sa) cout << " [skipped " << sa << "]";
                          if (ma > 1) cout << " [merged " << ma << "]";
                          cout << endl;
                          cout << "points: " << game->get_cplayer_info().global_obj("player")->get_int("minerals") << endl; 
                        }
			if (state->gfxm) {


				// don't draw if we are far behind
				if (abs(vf-af) < 10) {
					state->gfxm->draw();
					state->just_drew = true;
				}
			}else if (state->gui){
				state->gui->event();
				state->gui->display();
				if (state->gui->quit) exit(0);
			}

			return true;
		}
	}

	return false;
}

void AIEventHandler::updateWorld(const GameChanges* changes){
	assert(changes);
	AIObject* aiob;	

	const Game *game = &HANDLER->state->gsm->get_game();
  	const sint4 cid = game->get_client_player(); 
	player_bitmap = 1 << cid;
	neutral_bitmap = 1 << game->get_player_num();
	enemy_bitmap = ( neutral_bitmap - 1) ^ player_bitmap;
	all_bitmap = player_bitmap | neutral_bitmap | enemy_bitmap;

//	cout << bitset<32>(player_bitmap)<<" "<<bitset<32>(enemy_bitmap)<<" "<<bitset<32>(all_bitmap)<<endl;

//	Game* game=&state->gsm->get_game();

//	cout << "---AIEventHandler::updateWorld---" << endl;
	//cout << "adding new Obstacles:" << endl;
	FORALL(changes->new_boundaries, it){
		collector->addAIObject(new Obstacle( (*it)->get_GameObj() ));
	}
	//cout << endl << "adding new GameObjs" << endl;
	FORALL( changes->new_objs, it ){
		GameObj *gob = (*it)->get_GameObj();
		if (!gob) 
			continue; // we are interested only in GameObjs
		if (!gob->sod.in_game) 
			continue;
		//cout << "owner==" << *gob->sod.owner << ", gob->name=" << gob->bp_name() << endl;
		/*if (*gob->sod.owner != game->get_client_player())
		  continue;*/
		aiob = AIObjectCollector::createAIObject(gob);
		if (aiob) {
			collector->addAIObject(aiob);
		        //cout << "Added object of id:"<<aiob->getId()<<endl;
                }
	}


	FORALL( changes->vanished_objs, it ){
		GameObj *gob = (*it)->get_GameObj();
		if (!gob) 
			continue; // we are interested only in GameObjs
		if (!gob->sod.in_game) 
			continue;
		collector->removeGameObj(gob);
	}
	bool build_new_graph=false;
	FORALL( changes->dead_objs, it){
		GameObj *gob = (*it)->get_GameObj();
		if (!gob) 
			continue; // we are interested only in GameObjs
		if (!gob->sod.in_game) 
		{
			cout << "dead obj not in game:" << gob->bp_name() << endl;
			continue;
		}
		cout << "dead obj in game:" << gob->bp_name() << endl;
		collector->removeGameObj(gob);
		if (gob->bp_name()=="mineral"){
			build_new_graph=true;
		}


	}
	FOR_EACH_OBJ(o, AIObject) {
	  // cout << "Updating attributes of "<<o->gob->bp_name()<< ": "<<o->getId()<<endl;
		o->cacheAttributes();
                Worker* w = dynamic_cast<Worker*>(o);
                if (w){
                  w->updateInactivity();
                  w->updateMineral();
                }
	}

	if (build_new_graph){
		Configuration graph_cfg=config->getConfig("graph");
		Graph* new_graph = Graph::generate(&graph_cfg);
		REP(i,graph->num_v) if (graph->inactive[i]) new_graph->inactive[new_graph->getVertex(graph->getLoc(i))]=true;
		delete graph;
		graph=new_graph;
	}

	if (!graph){
		cout << "create new graph" << endl;
		Configuration graph_cfg=config->getConfig("graph");
		graph = Graph::generate(&graph_cfg);

	}
	else 
		graph->update(changes);

}



void AIEventHandler::compute_actions(){
	const GameChanges* changes = &state->gsm->get_changes();

// Troche dziwnie.  Powinno byc chyba if (!coordinator)
// Zarzyk : !coordinator jest rownowazne !graph
// Ten komentarz mozna usunac.
	if (!coordinator){ // game initialization ( equivalent to if (!graph)
		coordinator = new Coordinator();
	}
	updateWorld(changes);
	++current_frame;
	bool ptest;
	Options::get("-pathfind_test",ptest);
	if (ptest){
		cout << "Running pathfinding test"<<endl;
		REP(i,50){
			Path p(Loc(10,10),Loc(300,300));
			cout <<i<<' ';
		}
		cout <<endl;
	} else
	coordinator->setAction();
}
