/* Headerfile for tournament manager software.
 *
 * June 7, 2006
 */

#ifndef TMANAGER_H
#define TMANAGER_H

#include <pthread.h>
#include <string>
#include <map>
#include <cstdio>
#include <fstream>
#include <cstdlib>
#include <sstream>
#include <sys/time.h>
#include <ctime>
#include <iostream>

using namespace std;

#define true      1
#define false     0
#define FOR(i,n)  for (i=0; i < (n); i++)
#define FOREVER   for (;;)
#define NO_RES 0        //For tournament matrix display: 
                        //0-only display result, 1- display everything other than result


enum result_flag_t { queued, pending, done };

struct PLAYERDATA
{
  string name;                 //player name
  string username;             //player's user name 
  int index;                   //player index
};

struct COMPUTER
{
  string name;                 //computer name
  int index;                   //computer index
  int taskCount;               //number of processes on this computer (can have more than 1 process on a server)
  bool isServer;               //true <=> computer is a server
  bool isUp;                   //true <=> computer is up and ok to use
};

struct GAME
{       
  int id;                      //game index
  int gameType;                //type of the game (1,2,or 3)
  result_flag_t progress;      //indicates whether the game is queued, pending, or done 
  string result[3];            //result strings -- 0 is server, 1 is player1, 2 is player2                       
  int seed;                    //game (map) seed
  PLAYERDATA player_list[2];   //players of the game): 1 element for game type 1, 2 elements for types 2 & 3
  COMPUTER *clients[2];        //client computers for this game: 1 client for game type 1, 2 for types 2 & 3
  COMPUTER *server;            //the server computer for this game
  timeval start_time;          //timestamp of the start of this game              
  timeval end_times[3];        //timestamp of the end of each thread 
  bool saved;                  //true <=> this game's results are saved in the config file
  pthread_t thread_srv;        //server thread
  pthread_t threads_ply[2];    //player threads
};

struct GAME_ARG
{
  GAME *game;
  int playerIndex;
};

//GLOBAL GAME DATA STRUCTURES:
extern map<int,GAME> game_queue;        //list of games to be played

//COMPUTERS:
extern map<int, COMPUTER> computers;    //indexed starting from 0

//GLOBAL PLAYERS DATA STRUCTURE:
extern map<int, PLAYERDATA> players;    //indexed starting from 0

//THREAD VARIABLES:            
extern pthread_t *threads_game;         //indexed using game id
extern pthread_mutex_t mutexGames;    


//Helper Functions:

/**
 * Error reporting function, which displays the given error
 * message to standard error and terminates the program with code 20.
 *
 * @param msg  error message to be displayed
 */
void Error(string msg);

/**
 * Helper function for parsing the configuration file which
 * skips sections of comments (denoted by #).
 * 
 * @param in  the input file stream (configuration file)
 * @return    a bool flag indicating whether or not end-of-file is reached (true - eof, false - otherwise)
 */
bool skipComments(ifstream &in);

/**
 * Displays the help message to standard out when -help is the command line argument.
 * Exits the program with code 10.
 */
void print_help();

/**
 * Displays all players' data to standard out, as read in from the configuration file.
 *
 * @param single          flag indicating whether one or two games are played per map (true-one game, false-two games)
 * @param games_per_pair  the number of games to be played per pair of players (in a given round)
 * @param playerCount     the total number of players 
 */
void displayPlayers(bool single, int games_per_pair, int playerCount);

/**
 * Displays all games in the queue. Shows game index, player1 name, player2 name
 * @param gameCount  the number of games in the queue
 * @param gameType   the type of game(1,2,or3)
 */
void displayGames(int gameCount, int gameType);


//Central Functions:

/**
 * Generates the game queue.
 * 
 * @param playerCount     the total number of players
 * @param games_per_pair  the number of games to be played per pair of players (in a given round)
 * @param single          flag indicating whether one or two games are played per map (true-one game, false-two games)
 * @param game_seed       base map seed found by XORing all players' seeds 
 * @param gameType    the type of game(1,2,or 3)
 * @return                total number of games in the queue 
 */
int buildGameList(int games_per_pair, bool single, int playerCount, int game_seed, int gameType);

/**
 * Makes the ssh call to a client and stores the result of this call 
 * as this client's result for this particular game. (not fully implemented yet)
 * Writes all received results to disk files.
 *
 * @param arg  pointer to the game that is being played
 */
void *playClient(void *arg);

/**
 * Makes the ssh call to the server and stores the result of this call 
 * as the server's result for this particular game. (not fully implemented yet)
 * Writes all received results to disk files.
 * 
 * @param arg  pointer to the game that is being played
 */
void *playServer(void *arg);

/**
 * Sets up the server and clients and spawns the corresponding threads.
 *
 * @param arg         pointer to the game that is to be played 
 */
void *playGame(void *arg);


/**
 * Updates the contents of the configuration file by appending new game results to the existing file.
 *
 * @param filename        the name of the configuration file
 * @param playerCount     the total number of players
 * @param single          flag indicating one or two games per map 

 * @param gameType        the type of game(1,2,or 3)
 */
void updateConfig(string filename, bool single, int playerCount, int gameCount, int gameType);

/**
 * Checks pending games for completed ones.
 *
 * @param gameCount   the total number of games in game_queue 
 * @param game_length the expected length of a game
 * @param epsilon     the allowable grace period for a game
 * @return            indicates whether or not all games are done being played
 */
bool checkGames(int gameCount, double game_length, double epsilon);

/**
 * Checks for available computers.
 *
 * @param compCount       the total number of computers that were initially available 
 * @param  gameCount      the total number of games in game_queue
 * @return                Return true if we are done all the games
 */
bool checkComputers(int compCount, int gameCount);

/**
 * Returns the next queued game.
 *
 * @param  gameCount      the total number of games in game_queue
 * @return                the index of the next game in game_queue that is queued
 */
int getNextQueuedGame(int gameCount);

#endif
