/** @file An implementation of a simple k-means clustering method
          intended to cluster the minerals into tight patches.
*/

#include "Game3_accessories.H"

#include "KMlocal.h"  // k-means algorithms

#include <algorithm>
#include <numeric>
#include <math.h>

#include "boost/smart_ptr.hpp"
#include "boost/array.hpp"

using boost::shared_ptr;
using boost::scoped_array;
using boost::array;

#define TEST(x)
//#define TEST(x) x

/**************************************************************************/

typedef Vector< shared_ptr<KMfilterCenters> > KMCentersArray;
REGISTER_TYPEOF(4701, KMCentersArray::iterator);
REGISTER_TYPEOF(4702, KMCentersArray::const_iterator);

/**************************************************************************/

// Termination conditions for K-Means.
// Taken from kmlsample.cpp of the knlocal library.
static const KMterm term(
                100, 0, 0, 0,		// run for 100 stages
		0.10,			// min consec RDL
		0.10,			// min accum RDL
		3,			// max run stages
		0.50,			// init. prob. of acceptance
		10,			// temp. run length
		0.95);			// temp. reduction factor

/**************************************************************************/

static void point2point_distance(const KMdata& dataPts, double *res)
{
  const int n = dataPts.getNPts();
  for( int i=0 ; i < n-1 ; i++ )
    for( int j=i+1 ; j < n ; j++ ) {
      double dx = dataPts[i][0]-dataPts[j][0];
      double dy = dataPts[i][1]-dataPts[j][1];
      res[i*n + j] = res[j*n + i] = sqrt((dx*dx)+(dy*dy));
    }
  // zero out the main diagonal
  for( int i=0 ; i < n ; i++ )
    res[i*n + i] = 0;

  TEST(
    cout << "point2point_distance = [" << endl;
    for( int i=0 ; i < n ; i++ ) {
      for( int j=0 ; j < n ; j++ ) {
        cout << res[i*n + j] << " ";
      }
      cout << ";" << endl;
    }
    cout << "];" << endl;
  );
}

/**************************************************************************/

static void point2cluster_distance(const double *p2p_distance, const int nPts, KMctrIdxArray clusterOf, const int nClusters, double *res)
{
  // initialize res to zero
  fill(res, res+(nPts*nClusters), 0);
  // for each point, sum its distances to the points in each cluster
  for( int i = 0 ; i < nPts ; i++ ) {
    for( int j = 0 ; j < nPts ; j++ ) {
      res[i*nClusters + clusterOf[j]] += p2p_distance[i*nPts + j];
    }
  }
  // and average
  for( int c = 0 ; c < nClusters ; c++ ) {
    int pointsInCluster = count(clusterOf, clusterOf+nPts, c);
    for( int p = 0 ; p < nPts ; p++ ) {
      res[p*nClusters + c] *= 1.0/(pointsInCluster - (clusterOf[p]==c)? 1:0);
    }
  }

  TEST(
    cout << "point2cluster_distance = [" << endl;
    for( int p=0 ; p < nPts ; p++ ) {
      for( int c=0 ; c < nClusters ; c++ ) {
        cout << res[p*nClusters + c] << " ";
      }
      cout << ";" << endl;
    }
    cout << "];" << endl;
  );
}

/**************************************************************************/

static double compute_avg_silhouette(const double *p2p_distance, const int nPts, KMctrIdxArray clusterOf, const int nClusters)
{
  assert( nClusters >= 2 && "Silhouettes are not defined for less than 2 clusters");

  scoped_array<double> p2c_distance(new double[nPts*nClusters]);
  scoped_array<double> silhouettes(new double[nPts]);

  // for each point: compute average distance to the points in every cluster
  point2cluster_distance(p2p_distance, nPts, clusterOf, nClusters, p2c_distance.get());

  // compute the silhouette of each point
  for( int p = 0 ; p < nPts ; p++ ) {
    double inside_distance = p2c_distance[p*nClusters + clusterOf[p]];
    double distance_to_closest_other_cluster = (clusterOf[p]!=0)? p2c_distance[p*nClusters] : p2c_distance[p*nClusters + 1];
    for( int c = (clusterOf[p]!=0)? 1 : 2 ; c < nClusters ; c++ ) {
      if( c != clusterOf[p] && p2c_distance[p*nClusters + c] < distance_to_closest_other_cluster )
        distance_to_closest_other_cluster = p2c_distance[p*nClusters + c];
    }
    silhouettes[p] = (distance_to_closest_other_cluster-inside_distance) /
                     (max(distance_to_closest_other_cluster, inside_distance));
  }

  // and average
  return accumulate(silhouettes.get(), silhouettes.get()+nPts, 0.0) / nPts;
}

/**************************************************************************/

void simple_clustering(const Vector<ScalarPoint> &data, 
                        uint4 min_k, uint4 max_k, 
                        Vector<sint4> &result)
{
  const int n = data.size();
  const int kChoices = max_k - min_k + 1;

  TEST( cout << "data=[" << endl;
        FORALL(data, it) cout << it->x << "," << it->y << endl;
        cout << "];" << endl; );

  // Build the input for the k-means (KM) alg
  KMdata dataPts(2, n); // 2 dimensional world
  FORU(i, n) {
    dataPts[i][0] = data[i].x;
    dataPts[i][1] = data[i].y;
  }

  // build the KM filtering structure
  dataPts.buildKcTree();

  // output for different values of k, kept for choosing the best k.
  KMCentersArray centers(kChoices);

  // Run KM for several values of k
  for( int i = 0, k = min_k ; i < kChoices ; i++, k++ ) {
    // allocate centers
    shared_ptr<KMfilterCenters> p(new KMfilterCenters(k, dataPts));
    centers[i] = p;
    // Create the algorithm
    KMlocalHybrid kmHybrid(*centers[i], term);
    // and execute
    *centers[i] = kmHybrid.execute();
  }

  /* Select the best value for k using the maximum average silouhette.
     The silhouette is a measure of how close each point in one cluster 
     is to points in the neighboring clusters. This measure ranges from +1 
     for points that are very distant from neighboring clusters, 
     to -1, indicating points that are probably assigned to the wrong 
     cluster. A high average silhouette inidcates good clustering.
     Silhouette is defined as:
       S_i = (b_i - a_i) / max(b_i, a_i)
     where a_i is the average distance from the ith point to the other 
     points in its cluster, and b_i is the average distance from the 
     ith point to points in the closest neighboring cluster.
   */

  // compute point2point distances, used later to compute silhouettes
  scoped_array<double> p2p_distance(new double[n*n]);
  point2point_distance(dataPts, p2p_distance.get());

  // compute average silhouette for each k
  scoped_array<double> avg_silhouette(new double[kChoices]);
  for( int i = 0, k = min_k ; i < kChoices ; i++, k++ ) {
    // get cluster assigmnets
    scoped_array<KMctrIdx> closestCtr(new KMctrIdx[n]);
    centers[i]->getAssignments(closestCtr.get(), NULL);

    // compute silhouette
    avg_silhouette[i] = compute_avg_silhouette(p2p_distance.get(), n, closestCtr.get(), k);
  }

  TEST(
    cout << "avg_silhouette = [";
    for( int i = 0 ; i < kChoices ; i++ )
      cout << avg_silhouette[i] << " ";
    cout << "];" << endl;
  );

  // Choose the best k, according to the maximum average silhouette
  int bestK_idx = max_element(avg_silhouette.get(), avg_silhouette.get()+kChoices) -
                  avg_silhouette.get();

/*  // Choose the best k, according to minimum distortion
  int bestK = MIN_NUM_CLUSTERS;
  double minDistortion = centers[0]->getDist(false);

  for( int k = MIN_NUM_CLUSTERS+1 ; k <= MAX_NUM_CLUSTERS ; k++ ) {
    double dist = centers[k-MIN_NUM_CLUSTERS]->getDist(false);
    if( dist < minDistortion ) {
      minDistortion = dist;
      bestK = k;
    }
  }*/

  // create output cluster assignments
  scoped_array<KMctrIdx> closestCtr(new KMctrIdx[n]);
  centers[bestK_idx]->getAssignments(closestCtr.get(), NULL);
  result.clear(); result.reserve(n);
  FORU(i, n)
    result.push_back(closestCtr[i]);

  TEST(
    cout << "mineral2cluster = [" << endl;
    FORU(i, dataPts.getNPts()) {
      cout << dataPts[i][0] << "," << dataPts[i][1] << "," << closestCtr[i] << ";" << endl;
    }
    cout << "];" << endl;
  );
}
