/** @file Game.C
    @see Game.H

    $Id: Game.C,v 1.67 2007/02/12 02:46:34 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/Game.C,v $
*/

// This is an ORTS file
// (c) Michael Buro
// (c) Tim Furtak
// licensed under the GPL

#include "Global.H"
#include "Game.H"
#include "GUI.H"
#include "GameObj.H"
#include "GameObjFactory.H"
#include "CycleCounter.H"
#include "BoolMatrix.H"
#include "PlayerInfo.H"
#include "Token.H"
#include "GameTile.H"
#include "ScriptParser.H"
#include "Options.H"
#include "Blueprints.H"
#include "ByteData.H"
#include "FileArchive.H"

#include "ScriptVector.H"
#include "ScriptString.H"
#include "ScriptSet.H"

using namespace std;

uint8 t_move;
uint8 t_view;
uint8 t_cact;
uint8 t_eact;

#define PARSE_ERR(s) {\
  errstr << endl << ": parse error at line " << tokenizer.cur_line();\
  ERR(s);\
}

#define PARSE_ERR2(s1, s2) {\
  errstr << endl << ": parse error at line " << tokenizer.cur_line();\
  ERR2(s1,s2);\
}

//===================================================================

Game::Options::Options()
{
  ::Options::get("-nfog", nfog);
  ::Options::get("-seed", seed);
  //  ::Options::get("-packfile", pack_file);
  cpind = -1;
}

//-------------------------------------------------------------------

void Game::Options::add()
{
  ::Options o("Game");
  o.put("-packfile", string(""), "file containing compressed data");
  o.put("-seed",     sint4(0),   "game rng seed");

  Vector<string> v;
  o.put("-D", v, "script variables to define");
}

//===================================================================

// prevent copying and assigning (not implemented yet)
Game::Game(const Game &) : SaveLoadAbs() { ERR("what"); }
Game &Game::operator=(const Game &) { ERR("what"); return *this; }

//===================================================================

/** Clips tile x/y coordinates to the board and returns the index. */

sint4 Game::tile_index(sint4 x, sint4 y) const
{
  if (x < 0) x = 0;
  else if (x >= points_x) x = points_x-1;
  if (y < 0) y = 0;
  else if (y >= points_y) y = points_y-1;

  x /= tile_points;
  y /= tile_points;
  sint4 ind = y*pf.get_width() + x;
  if (ind < 0 || ind >= pf.get_num()) ERR("tile index out of range");
  return ind;
}

//===================================================================

/** Call destructors for dead objects and remove (dead) objects that
    are no longer referenced. */

void Game::garbage_collect_objects()
{
  set<ScriptObj*> dead;
  ScriptObj::collect_dead(dead); // call destructors for dead objects

  // Remove dead objects from each player's object list.
  // Could be more efficient.

  FORS (q, player_n+1) {

    ObjCont &obj = objs[q];
    ObjCont::iterator it = obj.begin();

    while (it != obj.end()) {
      if (dead.find(*it) != dead.end()) {
        it = obj.erase(it);
        continue;
      }
      ++it;
    }
  }

  ScriptObj::collect_destroyed(); // delete dead objects
}

//===================================================================

/** Squares a distance and adds an error factor to include diagonal tiles.
    Used to compare the distance between tiles; k=1 ==> 8 surrounding tiles. */

sint4 Game::ok_dist2(sint4 k)
{
  return ((10*k+7)*(10*k+7))/100;  // ~ (k+sqrt(2)/2)^2
}

//===================================================================

void Game::pick_game(Game *game)
{
  FunctionNode::game = game;
  MemberFunctionCall<Game>::owner = game;
}

//-------------------------------------------------------------------

Game::Game(const string &world, const Game::Options &opt_) :
  opt(opt_), move_actions(this), simple_actions(this)
{
  pick_game(this);
  
  Tokenizer::init_cmd_def(); // update from command line definitions

  if (opt.pack_file != "") FileArchive::read_archive(opt.pack_file);

  rand.srand(opt.seed);
  srand48(opt.seed);
  
  game_over = false;
  save_req = false;
  save_and_quit = 0;

  tick_count = 0;
  magic = 1; // must be != 0

  actions.push_back(&simple_actions);
  actions.push_back(&move_actions);

  if (opt.cpind >= 0) {
    ofstream out_file("test.map");
    if (!out_file.is_open()) ERR("error opening output file");
    out_file << world << endl;
    out_file.close();
  }

  //  this->world = opt.cpind < 0 ? remove_secret_info(world) : world;
  this->world = world;
  ByteStream bs(world);

  // look for world header information
  read_header(bs);

  // fixme: error checks
  player_n = bs.read_c_int4();
  gtiles_x = bs.read_c_int4();
  gtiles_y = bs.read_c_int4();
  z_sight  = bs.read_c_int4();

  {
    sint4 temp = bs.read_c_int4();
    rotated = temp == 1;
  }

  sint4 tiles_x, tiles_y;
  
  if (rotated) {
    tiles_x = gtiles_x + gtiles_y;
    tiles_y = gtiles_x + gtiles_y;
  } else {
    tiles_x = gtiles_x;
    tiles_y = gtiles_y;
  }

  pf.allocate(tiles_x, tiles_y);
  
  tile_points = GameConst::TILE_POINTS; // for now just a copy of a constant
  inv_tile_points = 1.0/tile_points;

  points_x = tiles_x * tile_points;
  points_y = tiles_y * tile_points;

  fine_grid_res_bits = 6;
  fine_grid_res      = 1 << fine_grid_res_bits;

  sec_wh = fine_grid(48);  // was 32 (too slow for 512x512)
  cout << "sec_wh = " << sec_wh << endl;

  fine_points_x = fine_grid(points_x);
  fine_points_y = fine_grid(points_y);
  sec_nx        = (fine_points_x+sec_wh-1)/sec_wh;
  sec_ny        = (fine_points_y+sec_wh-1)/sec_wh;
  sec_num       = sec_nx * sec_ny;

  cout << sec_num << " motion sectors" << endl;

  objs = new Vector<ScriptObj::Ptr>[player_n+1];

  string token;
  Tokenizer tokenizer;
  tokenizer.init(bs.ptr(), bs.length(), "Game Info");

  token = tokenizer.next_tag();

  pf.clear();

  if (token == "<TILES>") {
    ByteStream bs(tokenizer.data_ptr, tokenizer.data_end);
    pf.read(bs);
    tokenizer.data_ptr = bs.ptr();
    token = tokenizer.next_tag();
    if (token != "</TILES>") ERR2("</TILES> expected", token.c_str());

  } else {
    if (token != "") tokenizer.putback_tag(token);
  }

  FORS (i, player_n+1) { // "+1" for NPC data
    
    players.push_back(new PlayerInfo(tiles_x, tiles_y, opt.cpind < 0));
    players[i]->my_objs = &objs[i];
    players[i]->gof.game = this;

    FORS (j, pf.get_num()) {

      Tile::Type t1, t2;
      pf(j).get_type(t1, t2);

      // mark unknown tiles as explored so they won't be sent
      if (t1 == Tile::UNKNOWN && t2 == Tile::UNKNOWN) {
        if (opt.cpind < 0)
          players[i]->explored.set(j % tiles_x, j / tiles_x);
      }
    }
  }

  NPC = players[player_n];

  cplayer = opt.cpind >= 0 ? players[opt.cpind] : 0;

  // create shared view and friend status matrices

  x_allied_with_y = new BoolMatrix(player_n+1, player_n+1); // player_n = NPC
  
  FORS (i, player_n+1) {
    FORS (j, player_n+1) {
      x_allied_with_y->set(i, j, i == j);
    }
  }

  init_motion_sectors();

  allocate_vision_memory();
  precompute_vision_data();

  MapView::b2t = bitmap2tind;
  MapView::t2b = tind2bitmap;


  add_script_functions();
  add_script_classes();

  //----World description----------------------

  tokenizer.set_line(0); // there's still a newline waiting
  while(tokenizer.good()) {

    token = tokenizer.next_tag();

    if      (token == "<DEFINE>")     read_define_block(tokenizer);
    else if (token == "<ACTIONS>")    read_action_block(tokenizer);
    else if (token == "<BLUEPRINTS>") read_blueprint_block(tokenizer);
    else if (token == "<OBJECTS>")    read_object_block(tokenizer);
    else if (token == "<TILEDEF>")    read_tiledef_block(tokenizer);

    else {
      token = tokenizer.next_token();
      if (token != "") {
        cout << "type = " << tokenizer.token_type() << endl;
	ERR2("expecting a tag, found:", token.c_str());
      }
      tokenizer.putback(token);
    }
  }

  FORALL (players, it) {
    (*it)->gof.lock(); // the blueprints and actions for the game description
  }
  GameObjBlueprint::all_locked = true;
  //-------------------------------------------

  generate_boundaries();

  FORALL (players, it) {
    ScriptObj *s = (*it)->global_obj("game_state");
    if (!s) continue;
    s->set_int("w", points_x, false);
    s->set_int("h", points_y, false);
  }
}

//===================================================================

void Game::read_blueprint_block(Tokenizer &tokenizer, bool client_extend)
{
  string token;

  while (tokenizer.good()) {

    token = tokenizer.next_tag();
    if (token == "</BLUEPRINTS>") break;
    if (token != "") {
      cout << "found: " <<  token << endl;
      ERR ("</BLUEPRINTS> expected");
    }
    token = tokenizer.next_token();

    if (token == "global" || token == "shared") {
      
      bool is_global = token == "global";
      bind();

      string class_name = tokenizer.next_token();
      if (tokenizer.token_type() != Tokenizer::ID) PARSE_ERR("class name expected");
      
      FOREVER {
        string object_name = tokenizer.next_token();
        if (tokenizer.token_type() != Tokenizer::ID) PARSE_ERR("object name expected");
        
        FORU (i, players.size()) {

	  PlayerInfo &pi = *players[i];

          // create the global object

          if (opt.load_file != "") continue;

          ScriptObj **obj = new ScriptObj*;
          ScriptObj *sob;

          if (is_global) {
            
            sob = pi.gof.create(class_name);
            *obj = sob;
            pi.new_global_obj(object_name, obj);

          } else { // shared object
            if (i == 0) {
              sob = pi.gof.create(class_name);
              *obj = sob;
              
              cout << "adding shared object " << object_name << endl;
              pi.new_shared_obj(object_name, obj);

            } else {
              sob = pi.shared_obj(object_name); // get the shared object
              assert(sob);
              *obj = sob;
            }
          }

          GameObj *gob = sob->get_GameObj();
          if (gob != 0) {
            gob->init_sod();
            gob->sod.is_global = true;
            gob->sod.is_shared = !is_global;
          }

          if (!client_extend) pi.new_id(*obj);
          objs[i].push_back(*obj);
          
          if (sob->has_attr("owner")) sob->set_int("owner", i, true);
          sob->add_ref();     // make sure that the object is never destroyed

        }
        
        token = tokenizer.next_token();
        if (token == ";") break;
        if (token != ",") PARSE_ERR("';' or ',' expected");
      }
      
      continue;
    } // end of global/shared object

    // either a class or a blueprint
    /* Practically, a blueprint is a class that can be used to create a parent object which
       may or may not include or contain classes/blueprints as sub-objects. */
    bool is_blueprint = token == "blueprint";
    bool is_class     = token == "class";
    if (!is_blueprint && !is_class) {
      cout << "found " << token << flush;
      PARSE_ERR("Expecting 'blueprint' or 'class'");
    }
    
    string bname = tokenizer.next_token();

    // The client shouldn't execute any of the (default) constructors or destructors,
    // only the server should care about them.
    bool read_constructors = (get_client_player() < 0) || client_extend;

    if (client_extend) { // try to find the blueprint if it already exists

      GameObjFactory::Name2BlueprintMap::iterator it;

      tokenizer.push_bookmark();

      FORALL (players, q) {
        
        GameObjFactory::Name2BlueprintMap *n2b;
        GameObjFactory::Name2ConstructorMap *n2c;
        n2b = &(*q)->gof.name2blueprint;
        n2c = &(*q)->gof.name2constructor;

        tokenizer.flip_bookmark();
        it = n2b->find(bname);
        if (it != n2b->end()) {
          // extend an existing blueprint
          it->second->name = bname;
          it->second->parse_class_def(tokenizer, *n2b, *n2c, *this, **q, read_constructors);
        } else {
          // create a new blueprint
          GameObjBlueprint bp(bname, is_blueprint); // can_instance <- is_blueprint
          bp.parse_class_def(tokenizer, *n2b, *n2c, *this, **q, read_constructors);
          
          (*q)->gof.register_blueprint(bname, bp);
        }
      }
      tokenizer.drop_bookmark();

    } else { // Part of the game description.
             // The blueprint shouldn't already exist, create a new one and register it.

      tokenizer.push_bookmark();

      FORALL (players, q) {
        
        GameObjFactory::Name2BlueprintMap *n2b;
        GameObjFactory::Name2ConstructorMap *n2c;
        n2b = &(*q)->gof.name2blueprint;
        n2c = &(*q)->gof.name2constructor;

        tokenizer.flip_bookmark();

        GameObjBlueprint bp(bname, is_blueprint); // can_instance <- is_blueprint

        bp.parse_class_def(tokenizer, *n2b, *n2c, *this, **q, read_constructors);
        
        (*q)->gof.register_blueprint(bname, bp);
      }

      tokenizer.drop_bookmark();
    }
  }

  // If an action has an id associated with it it is set as outgoing (client use).
  // Therefore, if this is the server, clear the id to execute all actions.
  bind();
}

//===================================================================

void Game::bind()
{
  FORALL (players, it) {
    (*it)->gof.bind(*(*it), *this);
    (*it)->gof.lock();
    if (get_client_player() < 0) {
      FORALL ((*it)->gof.code2blueprint, it2) {
        it2->second->clear_action_map();
      }
    }
  }
}

//===================================================================

void Game::read_action_block(Tokenizer &tokenizer)
{
  string token;
  while (tokenizer.good()) {
    token = tokenizer.next_tag();
    if (token == "</ACTIONS>") break;
    if (token != "") ERR ("</ACTIONS> expected");
    
    ObjActionDefinition actdef;
    actdef.name = tokenizer.next_token();
    actdef.parse_action_def(tokenizer);
    actdef.act_type = ObjActionDefinition::SCRIPT;

    actdef.file_name = tokenizer.file_name();
    actdef.line_num = tokenizer.cur_line();
    
    GameObjBlueprint::ext_action_defs.push_back(actdef);
  }
  
  FORALL (players, it) (*it)->gof.bind(*(*it), *this);
}

//===================================================================

void Game::read_define_block(Tokenizer &tokenizer)
{
  string token;
  while (tokenizer.good()) {
    token = tokenizer.next_tag();
    if (token == "</DEFINE>") break;
    if (token != "") ERR ("</DEFINE> expected");
    string var = tokenizer.next_token(false); // ignore redefines
    sint4 f_line = tokenizer.cur_line();
    
    string def = tokenizer.next_token();
    if (tokenizer.cur_line() != f_line || def == "") {
      ERR2 ("expecting definition for", var.c_str());
    }
    
    // Read until a new line.  Conveniently ignore trailing whitespace.
    FOREVER {
      token = tokenizer.next_raw_token(false);
      if (tokenizer.cur_line() != f_line) {
        tokenizer.putback(token);
        break;
      }
      if (token == "") break;
      if (tokenizer.token_type() != Tokenizer::COMMENT)
        def += token;
    }
    tokenizer.define_var(var, def);
  }
}

//===================================================================

void Game::read_object_block(Tokenizer &tokenizer)
{
  string token;

  FORALL (players, q) {
    (*q)->gof.bind(**q, *this);
    (*q)->gof.lock();
  }
  
  while (tokenizer.good()) {

    token = tokenizer.next_tag();
    if (token == "</OBJECTS>") break;
    if (token != "") ERR("expecting </OBJECTS>");
    token = tokenizer.next_token();
    
    string obj_type = token;
    ScriptObj *pobj = 0;
    
    FOREVER {
      
      token = tokenizer.next_tag();

      if (token == "<INT_ATTR>") {
        string attr;
        sint4 val;
        // get the owner (must be first), and create an object
        if (pobj == 0) {
          token = tokenizer.next_token();
          if (token != "owner")
	    ERR2 ("'owner' must be the first attribute listed : ", token.c_str());

          val = atoi(tokenizer.next_token().c_str());

          if (val >= 0 && val < player_n) {
            pobj = players[val]->gof.create(obj_type);
            objs[val].push_back(pobj);
          } else {
            pobj = NPC->gof.create(obj_type); // invalid owner -> NPC
            objs[player_n].push_back(pobj);
            val = player_n;
          }
          pobj->set_int(token, val);
        }
        
        FOREVER {
          token = tokenizer.next_tag();
          if (token == "</INT_ATTR>") break;
          if (token != "") ERR("expecting </INT_ATTR>");
          token = tokenizer.next_token();
          
          attr = token;
          val = atoi(tokenizer.next_token().c_str());
          pobj->set_int(attr, val);
        }
      }
      break;
    }

    GameObj *temp = pobj->get_GameObj();
    if (temp != 0) temp->init_sod();
  }
}

//===================================================================

void Game::read_tiledef_block(Tokenizer &tokenizer)
{
  string token;
  sint4 count;

  count = atoi(tokenizer.next_token().c_str());
  num_tile_types = count+1; // include unknown for convenience

  FORU (i, count+1) {
    FORU (j, count+1) {
      if (i == 0 || j == 0) {
        tile_collision_mask.push_back(0);
        continue;
      }
      sint4 v = atoi(tokenizer.next_token().c_str());
      tile_collision_mask.push_back(v);
    }
  }

  tile_height_diff_mask = atoi(tokenizer.next_token().c_str());
  
  token = tokenizer.next_tag();
  if (token != "</TILEDEF>") ERR ("</TILEDEF> expected");
}

//===================================================================

Game::~Game()
{
  deallocate_vision_memory();

  ObjAction::clear();

  FORU (i, player_n+1) {
    FORALL (objs[i], j) delete *j;
  }
  delete [] objs; objs = 0;
  
  FORALL (boundaries, i) { delete *i; *i = 0; }
  delete x_allied_with_y; x_allied_with_y = 0;
  FORS (i, player_n+1) { delete players[i]; players[i] = 0; }
}

//===================================================================

/** Check for any special op-codes in the client(s) action strings. */

void Game::preprocess_actions(Vector<string> &actions)
{
  FORALL (actions, it) {
    const string &s = *it;
    if (s.empty()) continue;
    ByteStream bs(s);
    sint4 id = bs.read_b_int4();
    if (id == -1) save_req = true;
  }
}

//===================================================================

bool Game::simulation_step(Vector<string> &actions, Vector<string> &views)
{
  tick_count++;

  CycleCounter c1;
  exec_actions(actions);   CycleCounter c2; t_eact += c2.count() - c1.count();
  move_objects();          CycleCounter c3; t_move += c3.count() - c2.count();
  gen_view_diffs(views);   CycleCounter c4; t_view += c4.count() - c3.count();
  return true;
}

//===================================================================

string Game::get_client_world()
{
  string w = remove_secret_info(this->world);

  string header = "DEF ";

  Vector<string> def;
  ::Options::get("-D", def);
  FORALL (def, i) {
    header += *i + " ";
  }
  header += "\n";

  return header + w;
}

//-------------------------------------------------------------------

void Game::read_header(ByteStream &bs)
{
  Tokenizer t;
  t.init(bs.ptr(), bs.length());

  string token, line;

  do {
    token = t.next_raw_token(false);
  } while (t.token_type() == Tokenizer::SPACE);

  if (token != "DEF") return;

  //------------------------------

  line = t.read_line();

  Vector<string> var;
  Vector<string> def;

  string d;
  FORALL (line, i) {
    if (*i == ' ') {
      if (d != "") {
        var.push_back(d);
        d = "";
      }
    } else {
      d += *i;
    }
  }

  FORALL (var, i) {
    const string &s = *i;
    size_t p = s.find("=");
    if (p == string::npos) {
      *i = s;
      def.push_back("");
    } else {
      string v = s.substr(0,p);
      string d = s.substr(p+1, s.size()-1);
      *i = v;
      def.push_back(d);
    }
  }

  Tokenizer::init_cmd_def(var, def);

  //------------------------------

  bs.set_ptr(t.data_ptr);

  return;
}

//===================================================================

/** Remove \<TILES\>...\</TILES\> and \<OBJECTS\>...\</OBJECTS\> from world description. */

string Game::remove_secret_info(const string &world)
{
  ostringstream os;
  istringstream is(world);
  bool skip = false;

  FOREVER {

    string t;
    getline(is, t);
    if (!is) break;

    if (t == "<TILES>" || t == "<OBJECTS>") skip = true;
    else if (t == "</TILES>" || t == "</OBJECTS>") skip = false;
    else {
      if (!skip) os << t << endl;
    }
  }

  return os.str();
}

//===================================================================

/** Clear the contents of the motion sectors for each collision layer. */

void Game::init_motion_sectors()
{
  fringe_secs.clear();
  boundary_secs.clear();
  FORALL (boundaries, i) delete *i;
  boundaries.clear();

  motion_secs.clear();

  fringe_secs.init(sec_num);
  boundary_secs.init(sec_num);
  motion_secs.init(sec_num);
}

//===================================================================

/** Encode the action to be taken by each object. */

void Game::gen_action_message(sint4 action_frame, string &msg) const
{
  msg.clear();

  if (save_req) {
    write_b_int4(msg, -1); // -1 is an invalid object id - use it for a save request
  }

  Vector<ScriptObj::Ptr> cobjs; //&cobjs = objs[opt.cpind]; // client object vector
  cplayer->objs(cobjs); // sorted ids from the id factory

  FORALL (cobjs, i) {
    GameObj *g = (*i)->get_GameObj();
    if (!g) cout << "not a game obj?" << endl;
    if (g && !g->is_dead()) {
      ObjId k = cplayer->get_id(g);
      string action = g->write_action(k);
      msg += action;
      g->queue_action(action_frame);
      g->clear_action();
    }
  }
}

//===================================================================

void Game::save(ostream &os) const
{
  OS_WRITE(player_n);
  OS_WRITE(tick_count);
  OS_WRITE(view_frame);

  // save objects
  FORS (i, player_n+1) {
    ObjCont &oc = objs[i];
    sint4 s = objs[i].size();
    OS_WRITE(s);

    FORS (j, s) {
      OS_WRITE(oc[j]); // save ScriptObj::Ptr
      ScriptObj &so = *oc[j];
      ObjId uid = ScriptObj::get_obj_id(&so);
      OS_WRITE(uid);
      uint4 code = so.code();
      OS_WRITE(code);
      so.save(os);
    }

    // save player info
    players[i]->save(os);
  }

  // save actions
  simple_actions.save(os);

  // save map
  {
    sint4 w = pf.get_width();
    sint4 h = pf.get_height();
    OS_WRITE(w);
    OS_WRITE(h);
    pf.save(os);
  }
}

//-------------------------------------------------------------------

void Game::load(istream &is)
{
  sint4 pn;
  IS_READ(pn);
  if (pn != player_n) ERR("player sizes don't match");

  IS_READ(tick_count);
  IS_READ(view_frame);

  // load objects
  FORS (i, player_n+1) {
    ObjCont &oc = objs[i];

    sint4 s;
    IS_READ(s);
    oc.clear();

    Vector<ScriptObj*> objs;
    Vector<ObjId> ids;

    FORS (j, s) {
      ScriptObj::Ptr p(0);
      IS_READ(p);
      oc.push_back(p);
      ObjId uid;
      IS_READ(uid);
      uint4 code;
      IS_READ(code);

      ScriptObj *so = players[i]->gof.create(code);
      so->load(is);

      GameObj *gob = so->get_GameObj();
      if (gob != 0) {
        update_pos(*gob);
        gob->init_sod();
        sint4 i = gob->get_int("owner");
        FORS (j, player_n) {
          if (x_allied_with_y->get(i,j))
            gob->sod.seen_by.set(j);
        }
      }

      objs.push_back(so);
      ids.push_back(uid);
    }

    ScriptObj::set_ids(objs, ids);

    // load player info (after ids have been loaded)
    players[i]->load(is);
  }

  // load actions
  simple_actions.load(is);

  // load map
  {
    sint4 w;
    sint4 h;
    IS_READ(w);
    IS_READ(h);
    if (w != pf.get_width()) ERR("map width doesn't match");
    if (h != pf.get_height()) ERR("map height doesn't match");
    pf.load(is);
  }
  generate_boundaries();
}

//===================================================================

void Game::add_script_functions()
{
  // compiled actions
  /*
  ScriptFunctionDef q;
  q.name = "CompiledAction1";
  q.num_obj_ids = 0;
  q.num_params = 0;
  ScriptParser::add(q, CompiledAction1);
  */

  //------------------------------
  // script interface functions
  //------------------------------

  // fixme: get rid of these; do them in the script
  ScriptParser::add("enable",           1,  0, script_GOB_Enable);
  ScriptParser::add("disable",          1,  0, script_GOB_Disable);
  ScriptParser::add("enabled",          1,  0, script_GOB_Is_Enabled);


  ScriptParser::add("printnum",         0,  1, script_printnum);
  ScriptParser::add("cout",             0, -1, &Game::script_cout);

  ScriptParser::add("tick",             0,  0, &Game::script_Tick);

  ScriptParser::add("distance",         2,  0, script_Distance);
  ScriptParser::add("notify_attack",    2,  1, script_Notify_Attack);
  ScriptParser::add("kill",             1,  0, script_GOB_Kill);
  ScriptParser::add("alive",            1,  0, script_GOB_Is_Alive);
  ScriptParser::add("typeof",           1,  0, script_GOB_Type);
  ScriptParser::add("type2name",        0,  1, &Game::script_GOB_Type2Name);
  ScriptParser::add("visible_by",       1,  1, script_GOB_Visible_By);
  ScriptParser::add("cancel",           1,  1, script_GOB_Cancel_Action);
  ScriptParser::add("has_attr",         1,  1, script_GOB_Has_Attr);
  ScriptParser::add("get_attr",         1,  1, script_GOB_Get_Attr);
  ScriptParser::add("get_comp",         1,  1, script_GOB_Get_Comp);

  ScriptParser::add("eval",            -1, -1, &Game::script_eval);
  ScriptParser::add("close_to",         2,  1, &Game::script_close_to);

  ScriptParser::add("create",           0,  2, &Game::script_GOB_Create);
  ScriptParser::add("create_id",        0,  2, &Game::script_GOB_CreateId);
  ScriptParser::add("move",             1, -1, &Game::script_GOB_Move);
  ScriptParser::add("position",         1,  0, &Game::script_GOB_Fix_Position);
  ScriptParser::add("can_see",          2,  0, &Game::script_GOB_Can_See);
  ScriptParser::add("footprint",        1,  1, &Game::script_GOB_Footprint);

  ScriptParser::add("get_tile",         0, -1, &Game::script_get_tile_type);
  ScriptParser::add("set_tile",         0, -1, &Game::script_set_tile_type);

  ScriptParser::add("map_width",        0,  0, &Game::script_map_width);
  ScriptParser::add("map_height",       0,  0, &Game::script_map_height);

  ScriptParser::add("sqrt",             0,  1, &Game::script_sqrt);

  ScriptParser::add("end_game",         0,  1, &Game::script_end_game);
  ScriptParser::add("save_game",        0,  0, &Game::script_save_game);
}

void Game::add_script_classes()
{
  FORALL (players, it) {
    PlayerInfo &pi = **it;
    ScriptObjConstructor soc;

    soc = &ScriptVector::create;
    pi.gof.register_blueprint("Vector", soc, 0);

    soc = &ScriptString::create;
    pi.gof.register_blueprint("String", soc, 0);

    soc = &ScriptSet::create;
    pi.gof.register_blueprint("Set", soc, 0);
  }
}

//===================================================================



#if 0
void Game::line(sint4 x1, sint4 y1, sint4 x2, sint4 y2, sint4 type)
{
  sint4 deltax = abs(x2 - x1);    // The difference between the x's
  sint4 deltay = abs(y2 - y1);    // The difference between the y's
  sint4 x = x1;                   // Start x off at the first pixel
  sint4 y = y1;                   // Start y off at the first pixel
  sint4 xinc1, xinc2, yinc1, yinc2;
  sint4 den, num, numadd, numpixels;

  if (x2 >= x1) {               // The x-values are increasing
    xinc1 = 1; xinc2 = 1;
  } else {                      // The x-values are decreasing
    xinc1 = -1; xinc2 = -1;
  }

  if (y2 >= y1) {               // The y-values are increasing
    yinc1 = 1; yinc2 = 1;
  }  else {                     // The y-values are decreasing
    yinc1 = -1; yinc2 = -1;
  }

  if (deltax >= deltay) {       // There is at least one x-value for every y-value
    xinc1 = 0;                  // Don't change the x when numerator >= denominator
    yinc2 = 0;                  // Don't change the y for every iteration
    den = deltax;
    num = deltax / 2;
    numadd = deltay;
    numpixels = deltax;         // There are more x-values than y-values
  } else {                      // There is at least one y-value for every x-value
    xinc2 = 0;                  // Don't change the x for every iteration
    yinc1 = 0;                  // Don't change the y when numerator >= denominator
    den = deltay;
    num = deltay / 2;
    numadd = deltax;
    numpixels = deltay;         // There are more y-values than x-values
  }

  Tile *p = tiles + y * tiles_x + x;
  const sint4 d1 = yinc1 * tiles_x + xinc1;
  const sint4 d2 = yinc2 * tiles_x + xinc2;

  FORS (i, numpixels+1) {
    p->type = type;
    num += numadd;              // Increase the numerator by the top of the fraction
    if (num >= den) {           // Check if numerator >= denominator
      num -= den;               // Calculate the new numerator value
      p += d1;                  // Change location as appropriate
      p->type = type;           // !!!is that correct? should depend on num (diagonals!)
    }
    p += d2;                    // Change location as appropriate
  }
}
#endif
