// $Id: Replayer.C,v 1.5 2006/12/24 21:07:04 orts_furtak Exp $

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Global.H"
#include "Replayer.H"
#include "Game.H"
#include "GameChanges.H"
#include "CycleCounter.H"
#include "ShutDown.H"
#include "ByteData.H"
#include "Options.H"

using namespace std;

//===================================================================

Replayer::Options::Options()
{
  ::Options::get("-seed", seed);
  ::Options::get("-loadreplay", load_replay);
  ::Options::get("-ids", ids);
  ::Options::get("-nfog", nfog);
}

//-------------------------------------------------------------------

void Replayer::Options::add()
{
  ::Options o("Replayer");

  o.put("-tmax",    sint4(-1),    "max # of ticks (<0:inf)");
  o.put("-nfog", "no fog of war");

  o.put("-loadreplay", string(""), "replay actions from file");
  o.put("-ids", string(""),        "accepted player id:key pairs separated by ,");  
}

//===================================================================

void Replayer::send_messages(const Vector<std::string> &msgs)
{
  if (msgs.size() != clients.size()) ERR("wrong player count");

  FORU (i, clients.size()) {
    GameChanges changes;
    ActionChanges action_changes;

    Game::pick_game(clients[i].game);

    clients[i].game->apply_view_diff(msgs[i], changes, action_changes);
  }
}

//-------------------------------------------------------------------

void Replayer::load_messages(Vector<string> &msgs)
{
  if (clients.size() == 0) return;

  msgs.clear();
  msgs.resize(clients.size());
  uint4 num_msgs, msg_size;

  if (fread(&num_msgs, sizeof(num_msgs), 1, load_file) != 1) {
    if (feof(load_file)) {
      REM("Reached replay EOF\n");
      finish();
      return;
    }      
    ERR("replay file corrupt 1");
  }
  
  if (num_msgs != clients.size()) ERR("Illegal client size in saved file");

  FORU (i, num_msgs) {
    int result = fread(&msg_size, sizeof(msg_size), 1, load_file);

    //printf("Load message %d of %d, res=%d, size %d\n", i+1, num_msgs, result, msg_size);

    if ((result == 0) || (result == EOF)) {
      fclose(load_file); REM("Reached replay EOF\n"); break;
    }

    msgs[i].clear();

    if (msg_size > 0) {

      char *place = new char[msg_size];
      
      if (fread(place, sizeof(char), msg_size, load_file) != msg_size)
	ERR("replay file corrupt 2");
      
      if (feof(load_file)) {
	fclose(load_file); REM("Reached replay EOF\n"); break;
      }
      
      msgs[i].append(place, msg_size);
      delete [] place; place = 0;
    }
  }
}

//-------------------------------------------------------------------

Replayer::Replayer()
{
  server = 0;
  frame = -1;
  load_file = 0;
}

//-------------------------------------------------------------------

Replayer::~Replayer()
{
  if (server) {
    delete server;
    server = 0;
  }

  Vector<ClientData>::iterator it = clients.begin(), end = clients.end();
  for (; it != end; ++it) {
    if ((*it).game) 
      delete (*it).game;
  }
  clients.clear();

  finish();
}

//-------------------------------------------------------------------

void Replayer::init(const string &world, Replayer::Options &opt, Vector<string> &views)
{
  views.clear();
  if (is_good()) finish();

  string server_world = world;
  string file_name = opt.load_replay;

  file_name = sanitize_path(file_name);
  load_file = fopen(file_name.c_str(), "r");
  if (!load_file) ERR2("could not open replay file", file_name);
  
  server_world.clear();
  uint4 size;
  if (fread(&size, sizeof(size), 1, load_file) != 1)
    ERR("insufficient data in replay file");
  
  assert(size > 0);
  char *loaded_world = new char[size];
  if (fread(loaded_world, sizeof(*loaded_world), size, load_file) != size)
    ERR("Did not read correct world size.");
  
  server_world.append(loaded_world, size);
  delete [] loaded_world; loaded_world = 0;


  Game::Options go;
  go.nfog = opt.nfog;
  server = new Game(server_world, go);
  sint4 nplayers = server->get_player_num();

  string client_world = server->get_client_world();

  frame = 0;

  // connect clients ...

  cout << "OK. Send game information ... " << flush;

  FORU (i, nplayers) {
    ClientData cd;
    Game::Options go;
    go.cpind = i;
    cd.game = new Game(client_world, go);
    clients.push_back(cd);

    string u;
    write_b_int4(u, i);
    views.push_back(u + client_world);
  }

  cout << "OK. Run the game ..." << endl;
}

//-------------------------------------------------------------------

void Replayer::tick()
{
  cout << "replayer tick" << endl;

  if (!is_good()) return;

  // run the game

  Vector<string> actions, views;
  FORS (i, server->get_player_num()) {
    actions.push_back("");
    views.push_back("");
  }

  server->preprocess_actions(actions);
  
  cout << "foo" << endl;

  // one simulation step
  Game::pick_game(server);
  server->simulation_step(actions, views);
  
  cout << "replayer ticker" << endl;

  // prepend frame information
  
  Vector<ClientData>::iterator i = clients.begin(), end = clients.end();
  for (; i != end; ++i) {
    Game *game = (*i).game;
    string u;
    
    // also change corresponding code in View2.C
    
    sint4 a = frame;                          // current frame
    sint4 b = 0;                              // action executed in this frame
    sint4 c = -1;                             // -1, no action deleted
    sint4 d = 0;
    sint4 e = game->get_save_req() ? 1 : 0;   // message flag: save game
    sint4 f = 1;                              // send simulation rate
    
    write_b_int4(u, a);
    write_b_int4(u, b);
    write_b_int4(u, c);
    write_b_int4(u, d);
    write_b_int4(u, e);
    write_b_int4(u, f);      
    
    views[i-clients.begin()] = u + views[i-clients.begin()];
  }
  
  cout << "1" << endl;
  send_messages(views); // send views
  cout << "2" << endl;
  load_messages(actions);
  cout << "3" << endl;
  
  if (server->get_finished()) {
    finish();
    return;
  }
  
  frame++;
}

//-------------------------------------------------------------------

void Replayer::tick(Vector<string> &views)
{
  cout << frame << endl;

  if (!is_good()) return;

  // run the game

  Vector<string> actions;
  FORS (i, server->get_player_num()) {
    actions.push_back("");
    views.push_back("");
  }

  if (frame) load_messages(actions);
  if (!is_good()) return;

  server->preprocess_actions(actions);
  
  // one simulation step
  Game::pick_game(server);
  server->simulation_step(actions, views);

  // prepend frame information
  
  Vector<ClientData>::iterator i = clients.begin(), end = clients.end();
  for (; i != end; ++i) {
    string u;
    
    // also change corresponding code in View2.C
    /*    
    sint4 a = frame;                          // current frame
    sint4 b = 0;                              // action executed in this frame
    sint4 c = -1;                             // -1, no action deleted
    sint4 d = 0;
    sint4 e = 0;                              // message flag: save game
    sint4 f = 1;                              // send simulation rate

    write_b_int4(u, a);
    write_b_int4(u, b);
    write_b_int4(u, c);
    write_b_int4(u, d);
    write_b_int4(u, e);
    write_b_int4(u, f);      
    */
    views[i-clients.begin()] = u + views[i-clients.begin()];
  }
  
  if (server->get_finished()) {
    finish();
    return;
  }
  
  frame++;
}

//-------------------------------------------------------------------

void Replayer::finish()
{
  if (load_file) {
    fclose(load_file);
    load_file = 0;
  }
  frame = -1;
}

//-------------------------------------------------------------------

bool Replayer::is_good()
{
  return frame >= 0;
}

//===================================================================


//===================================================================

ViewReplayer::Options::Options()
{
  ::Options::get("-loadreplay", load_replay);
  ::Options::get("-nfog", nfog);
}

//-------------------------------------------------------------------

void ViewReplayer::Options::add()
{
  ::Options o("ViewReplayer");

  o.put("-nfog", "no fog of war");
  o.put("-loadreplay", string(""), "replay actions from file");
}

//===================================================================

ViewReplayer::ViewReplayer()
{
  game = 0;
  frame = -1;
  load_file = 0;
}

//-------------------------------------------------------------------

ViewReplayer::~ViewReplayer()
{
  if (game) {
    delete game;
    game = 0;
  }

  finish();
}

//-------------------------------------------------------------------

void ViewReplayer::init(ViewReplayer::Options &opt)
{
  if (is_good()) finish();

  const string &data_file = opt.load_replay;

  load_file = fopen(data_file.c_str(), "r");
  if (!load_file) ERR2("could not open file", data_file);

  int result;
  uint4 view_size;
  result = fread(&view_size, sizeof(view_size), 1, load_file);
  if ((result == 0) || (result == EOF)) {
    fclose(load_file); REM("Reached replay EOF\n");
    finish();
    return;
  }

  char *data = new char[view_size];
  result = fread(data, view_size, 1, load_file);
  if ((result == 0) || (result == EOF)) {
    fclose(load_file); REM("Reached replay EOF\n");
    finish();
    return;
  }

  string world(data, view_size);


  Game::Options go;
  go.nfog = opt.nfog;
  game = new Game(world, go);

  frame = 0;

  cout << "OK." << endl;
}

//-------------------------------------------------------------------

void ViewReplayer::tick()
{
  cout << "replayer tick" << endl;

  if (!is_good()) return;

  int result;

  uint4 view_size;
  result = fread(&view_size, sizeof(view_size), 1, load_file);
  if ((result == 0) || (result == EOF)) {
    fclose(load_file); REM("Reached replay EOF\n");
    finish();
    return;
  }

  char *data = new char[view_size];
  result = fread(data, view_size, 1, load_file);
  if ((result == 0) || (result == EOF)) {
    fclose(load_file); REM("Reached replay EOF\n");
    finish();
    delete [] data;
    return;
  }

  string sdata(data, view_size);
  delete [] data;

  GameChanges gc;
  ActionChanges ac;

  game->apply_view_diff(sdata, gc, ac);

  frame++;
}

//-------------------------------------------------------------------

void ViewReplayer::finish()
{
  if (load_file) {
    fclose(load_file);
    load_file = 0;
  }
  frame = -1;
}

//-------------------------------------------------------------------

bool ViewReplayer::is_good()
{
  return frame >= 0;
}

//===================================================================

