// $Id: GfxTerrainTile.C,v 1.4 2006/08/29 00:10:24 nathanie Exp $

// This is an ORTS file (c) Keith Yerex, licensed under the GPL

#include "GfxTerrainTile.H"
#include "GfxTerrain.H"
#include "Game.H"
#include "GameTile.H"
#include "Image.H"
#include "GfxModule.H"
#include "WaterTile.H"
#include "Random.H"

using namespace std;

//===================================================================

Vec3<real4> GfxTerrainTile::noise_scale1;
Vec3<real4> GfxTerrainTile::noise_scale2;
Image GfxTerrainTile::noise;
real4 GfxTerrainTile::sz = 1.0/8.0;
real4 GfxTerrainTile::eps = 1.0/512.0;
bool GfxTerrainTile::outline = false;

Vector<GfxTerrainTile::GeometryType> GfxTerrainTile::types;
ShaderProgram GfxTerrainTile::tile_shader;

//===================================================================

GfxTerrainTile::GfxTerrainTile()
{
  initialized = false;
  flattened[0] = 0;
  flattened[1] = 0;
  flattened[2] = 0;
  flattened[3] = 0;
  tx = 0;
  ty = 0;
}

//-------------------------------------------------------------------

void GfxTerrainTile::add_type(sint4 id, GfxTerrainTile::GeometryType t)
{
  if (id >= (sint4)types.size()) {
    types.resize(id+1);
  }
  types[id] = t;
}

//-------------------------------------------------------------------

bool GfxTerrainTile::load_noise(const string &filename, const Vec3<real4> &scale1, const Vec3<real4> &scale2)
{
  if (!noise.loadTGA(filename.c_str())) {
    cout << "Unable to load terrain noise texture " << filename << endl;
    return 0;
  }

  if (noise.bpp() != 1) {
    cout << "Terrain noise texture is not 1bpp grayscale " << filename << endl;
    return 0;
  }
  
  noise_scale1 = scale1;
  noise_scale2 = scale2;

  return 1;
}

//-------------------------------------------------------------------

real4 GfxTerrainTile::get_noise_at(real4 x, real4 y)
{
  if (noise.width() == 0 || noise.height() == 0) return 0.0;

  real4 px = noise_scale1.x * x;
  real4 py = noise_scale1.z * y;
  
  sint4 nx = (sint4)(floor(px)) % noise.width(); if (nx < 0) nx += noise.width();
  sint4 ny = (sint4)(floor(py)) % noise.height(); if (ny < 0) ny += noise.height();
  real4 fx = px-floor(px);
  real4 fy = py-floor(py);

#if 1
  real4 n1 = noise_scale1.y * (real4)(noise[noise.width()*ny+nx]-128);
  real4 n2 = noise_scale1.y * (real4)(noise[noise.width()*ny+(nx+1) % noise.width()]-128);
  real4 n3 = noise_scale1.y * (real4)(noise[noise.width()*((ny+1) % noise.height()) + nx]-128);
  real4 n4 = noise_scale1.y * (real4)(noise[noise.width()*((ny+1) % noise.height()) + (nx+1) % noise.width()]-128);
#else
  real4 n1 = noise_scale1.y * (real4)(noise[noise.width()*ny+nx]);
  real4 n2 = noise_scale1.y * (real4)(noise[noise.width()*ny+(nx+1) % noise.width()]);
  real4 n3 = noise_scale1.y * (real4)(noise[noise.width()*((ny+1) % noise.height()) + nx]);
  real4 n4 = noise_scale1.y * (real4)(noise[noise.width()*((ny+1) % noise.height()) + (nx+1) % noise.width()]);
#endif

  real4 n = n1*(1-fx)*(1-fy) + n2*fx*(1-fy) + n3*(1-fx)*fy + n4*fx*fy;
  //  n *= 0.5;

/*  ny = (sint4)(floor(noise_scale2.z*y)) % noise_h;
  nx = (sint4)(floor(noise_scale2.x*x)) % noise_w;
  n += noise_scale2.y*(real4)(noise[noise_w*ny+nx]-128); */
  return n;
}

//-------------------------------------------------------------------

void GfxTerrainTile::compute_tc()
{
  int u=0, v=2;
  if (trans) { u=2; v=0;}

  FORS (i, 4) {
    FORS (j, 4) {
      texcoord[i][j][0] = hf*(vert[i][j][u]-vert[0][0][u])*(sz-eps)+tt1[0] + eps*hf;
      texcoord[i][j][1] = vf*(vert[i][j][v]-vert[0][0][v])*(sz-eps)+tt1[1] + eps*vf;
    }
  }
}

//-------------------------------------------------------------------

/*     N

   y  0-1             0-1
   |  |\|  TB_SPLIT   |/|  BT_SPLIT
   V  2-3             2-3
      x ->
*/

/**
   Smooth interpolation between tile corner heights.
   Does not necessarily corespond to tile geometry.
*/

real4 GfxTerrainTile::height_at(real4 x, real4 y) const
{
  assert(0 <= x && x <= 1.0);
  assert(0 <= y && y <= 1.0);  
 
  real4 i = x * 2.0 - 1.0;
  real4 j = y * 2.0 - 1.0;

  i = (cbrt(i) + 1) / 2.0;
  j = (cbrt(j) + 1) / 2.0;

  real4 h1 = h[3] * i + h[2] * (1.0 - i);
  real4 h2 = h[1] * i + h[0] * (1.0 - i);

  real4 h = h1 * j + h2 * (1.0 - j);

  return h;
}

//-------------------------------------------------------------------

void GfxTerrainTile::set_edge(sint4 a, sint4 b, real4 width)
{
  if (h[a] == h[b]) width = 0.0;
  set_edge(a, b, width, 0.0);
}

//-------------------------------------------------------------------

/**
   Sets tile interpolation points along one edge defined by indices
   a and b.  Each point is distance "width" from the middle of the edge,
   and is lowered by "depth" from the default height given by height_at().
*/

void GfxTerrainTile::set_edge(sint4 a, sint4 b, real4 width, real4 depth)
{
  Vec3<real4> &va = vert[a][a];
  Vec3<real4> &vb = vert[b][b];

  Vec3<real4> &vab = vert[a][b];
  Vec3<real4> &vba = vert[b][a];

#if 0
  // bilinear interpolation
  vert[a][b] = va * (0.5 - width) + vb * (0.5 + width);
  vert[b][a] = va * (0.5 + width) + vb * (0.5 - width);
#else

  real4 i = 0.5 + width;
  vab.x = va.x * i + vb.x * (1-i);
  vab.z = va.z * i + vb.z * (1-i);

  vba.x = va.x * (1-i) + vb.x * i;
  vba.z = va.z * (1-i) + vb.z * i;

  vab.y = height_at(vab.x, vab.z);
  vba.y = height_at(vba.x, vba.z);
#endif

  vab.y -= depth;
  vba.y -= depth;
}

//-------------------------------------------------------------------

void w_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  if (j & 1) {
    --j;
  } else {
    ++j;
    if (i & 1) {
      --i;
    } else {
      ++i;
      --x;
      w_point(i,j,x,y);
    }
  }
}

void e_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  if (j & 1) {
    --j;
    if (i & 1) {
      --i;
      ++x;
      e_point(i,j,x,y);
    } else {
      ++i;
    }
  } else {
    ++j;
  }
}

void n_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  if (j & 2) {
    j -= 2;
  } else {
    j += 2;
    if (i & 2) {
      i -= 2;
    } else {
      i += 2;
      --y;
      n_point(i,j,x,y);
    }
  }
}

void s_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  if (j & 2) {
    j -= 2;
    if (i & 2) {
      i -= 2;
      ++y;
      s_point(i,j,x,y);
    } else {
      i += 2;
    }
  } else {
    j += 2;
  }
}

void sw_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  s_point(i,j,x,y);
  w_point(i,j,x,y);
}

void nw_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  n_point(i,j,x,y);
  w_point(i,j,x,y);
}

void se_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  s_point(i,j,x,y);
  e_point(i,j,x,y);
}

void ne_point(sint4 &i, sint4 &j, sint4 &x, sint4 &y)
{
  n_point(i,j,x,y);
  e_point(i,j,x,y);
}

//-------------------------------------------------------------------

void GfxTerrainTile::initverts()
{
  GeometryType tW = types[typeW];
  GeometryType tE = types[typeE];

  real4 miny = numeric_limits<real4>::max();
  real4 maxy = numeric_limits<real4>::min();
  //  real4 x = aabb_o.x;
  //  real4 y = aabb_o.z;
  //  const real4 a = 0.4;
  //  const real4 b = 0.3;

  // corner positions

  vert[0][0].x = 0.0;
  vert[0][0].y = h[0];
  vert[0][0].z = 0.0;

  vert[1][1].x = 1.0;
  vert[1][1].y = h[1];
  vert[1][1].z = 0.0;

  vert[2][2].x = 0.0;
  vert[2][2].y = h[2];
  vert[2][2].z = 1.0;

  vert[3][3].x = 1.0;
  vert[3][3].y = h[3];
  vert[3][3].z = 1.0;

  // adjust interpolation vertices for different tile types ---------

  // corner positions
  {
    const real4 depth = 0.6;
    GfxTerrainTile *a, *b, *c;
    bool w;

    if (tW == WATER) {

      if (split != Tile::TB_SPLIT || tE == WATER) {
        a = ter->tile(tx-1, ty);
        b = ter->tile(tx-1, ty-1);
        c = ter->tile(tx,   ty-1);
        w = true;
        if (!a || a->get_typeE() != WATER) w = false;
        if (!c || c->get_typeS() != WATER) w = false;
        if (b) {
          if (b->split == Tile::TB_SPLIT) {
            if (types[b->typeW] != WATER || types[b->typeE] != WATER) w = false;
          } else {
            if (types[b->typeE] != WATER) w = false;
          }
        } else w = false;
        if (w) {
          vert[0][0].y -= depth;
        }
      }
      
      //------------------------------
      
      if (split != Tile::BT_SPLIT || tE == WATER) {
        a = ter->tile(tx-1, ty);
        b = ter->tile(tx-1, ty+1);
        c = ter->tile(tx,   ty+1);
        w = true;
        if (!a || a->get_typeE() != WATER) w = false;
        if (!c || c->get_typeN() != WATER) w = false;
        if (b) {
          if (b->split == Tile::BT_SPLIT) {
            if (types[b->typeW] != WATER || types[b->typeE] != WATER) w = false;
          } else {
            if (types[b->typeE] != WATER) w = false;
          }
        } else w = false;
        if (w) {
          vert[2][2].y -= depth;
        }
      }
    }

    //------------------------------

    if (tE == WATER) {
      
      if (split != Tile::BT_SPLIT || tW == WATER) {
        a = ter->tile(tx+1, ty);
        b = ter->tile(tx+1, ty-1);
        c = ter->tile(tx,   ty-1);
        w = true;
        if (!a || a->get_typeW() != WATER) w = false;
        if (!c || c->get_typeS() != WATER) w = false;
        if (b) {
          if (b->split == Tile::BT_SPLIT) {
            if (types[b->typeW] != WATER || types[b->typeE] != WATER) w = false;
          } else {
            if (types[b->typeW] != WATER) w = false;
          }
        } else w = false;
        if (w) {
          vert[1][1].y -= depth;
        }
      }
      
      //------------------------------

      if (split != Tile::TB_SPLIT || tW == WATER) {
        a = ter->tile(tx+1, ty);
        b = ter->tile(tx+1, ty+1);
        c = ter->tile(tx,   ty+1);
        w = true;
        if (!a || a->get_typeW() != WATER) w = false;
        if (!c || c->get_typeN() != WATER) w = false;
        if (b) {
          if (b->split == Tile::TB_SPLIT) {
            if (types[b->typeW] != WATER || types[b->typeE] != WATER) w = false;
          } else {
            if (types[b->typeW] != WATER) w = false;
          }
        } else w = false;
        if (w) {
          vert[3][3].y -= depth;
        }
      }
    }
  }

  // edges between tiles (adds a trench for adjacent water tiles)

  {
    const real4 depth = 0.6;
    const real4 width = 0.3;  // for water - distance from the midline (0.5)
    const real4 cliff = 0.3; // if the heights are different, the distance the cliff extends from the midline

    GfxTerrainTile *t;
#if 1
    if (get_typeW() == WATER && (t = ter->tile(tx-1, ty)) && t->get_typeE() == WATER) set_edge(0, 2, width, depth);
    else                                                                              set_edge(0, 2, cliff);

    if (get_typeN() == WATER && (t = ter->tile(tx, ty-1)) && t->get_typeS() == WATER) set_edge(0, 1, width, depth);
    else                                                                              set_edge(0, 1, cliff);

    if (get_typeE() == WATER && (t = ter->tile(tx+1, ty)) && t->get_typeW() == WATER) set_edge(1, 3, width, depth);
    else                                                                              set_edge(1, 3, cliff);

    if (get_typeS() == WATER && (t = ter->tile(tx, ty+1)) && t->get_typeN() == WATER) set_edge(2, 3, width, depth);
    else                                                                              set_edge(2, 3, cliff);
#else
    set_edge(0, 2, cliff);
    set_edge(2, 3, cliff);
    set_edge(1, 3, cliff);
    set_edge(0, 1, cliff);
#endif
  }

  // interior points
  {
    Vec3<real4> center(0.5, (h[0] + h[1] + h[2] + h[3]) * 0.25, 0.5);
    
    const real4 width = 0.2; // width of basins for WATER tiles
    const real4 depth = 0.6; // water depth
    const real4 cliff = 0.35; // distance from the center to place cliff interpolation points

    vert[1][2] = vert[2][1] = vert[3][0] = vert[0][3] = center;

    //    vert[3][0].x = 1.0;
    //    vert[3][0].y = 1.0;
    //    vert[3][0].z = 1.0;

    if (h[0] != h[1] || h[2] != h[3] || split != Tile::NO_SPLIT) {
      vert[1][2].x += cliff;
      vert[3][0].x += cliff;
      vert[0][3].x -= cliff;
      vert[2][1].x -= cliff;
      has_cliffs = true;
    }

    if (h[0] != h[2] || h[1] != h[3] || split != Tile::NO_SPLIT) {
      vert[1][2].z -= cliff;
      vert[0][3].z -= cliff;
      vert[3][0].z += cliff;
      vert[2][1].z += cliff;
      has_cliffs = true;
    }

    if (split != Tile::TB_SPLIT) {
      if (h[0] == h[1] && h[0] == h[2]) vert[0][3] = vert[1][2] = vert[2][1] = center;
      if (h[3] == h[1] && h[3] == h[2]) vert[3][0] = vert[1][2] = vert[2][1] = center;
    }

    if (split != Tile::BT_SPLIT) {
      if (h[2] == h[0] && h[2] == h[3]) vert[2][1] = vert[3][0] = vert[0][3] = center;
      if (h[1] == h[0] && h[1] == h[3]) vert[1][2] = vert[3][0] = vert[0][3] = center;
    }

    if (tW == WATER || tE == WATER) has_cliffs = true;

    if (tW == WATER && tE == WATER) {
      vert[2][1].x = vert[0][3].x = 0.5 - width;
      vert[3][0].x = vert[1][2].x = 0.5 + width;
      vert[2][1].z = vert[3][0].z = 0.5 + width;
      vert[0][3].z = vert[1][2].z = 0.5 - width;
      sint4 a[] = {1,2,3,0};
      sint4 b[] = {2,1,0,3};
      FORU (i, 4) {
        Vec3<real4> &v = vert[a[i]][b[i]];
        v.y = height_at(v.x, v.z) - depth;
      }
    } else {
      if (tW == WATER) {
        if (split == Tile::BT_SPLIT) {
          Vec3<real4> &v = vert[0][3];
          v.x = 0.5 - width;
          v.z = 0.5 - width;
          v.y = height_at(v.x, v.z) - depth;
        } else {
          Vec3<real4> &v = vert[2][1];
          v.x = 0.5 - width;
          v.z = 0.5 + width;
          v.y = height_at(v.x, v.z) - depth;
        }
      }
      else if (tE == WATER) {
        if (split == Tile::BT_SPLIT) {
          Vec3<real4> &v = vert[3][0];
          v.x = 0.5 + width;
          v.z = 0.5 + width;
          v.y = height_at(v.x, v.z) - depth;
        } else {
          Vec3<real4> &v = vert[1][2];
          v.x = 0.5 + width;
          v.z = 0.5 - width;
          v.y = height_at(v.x, v.z) - depth;
        }
      } else {
        sint4 a[] = {1,2,3,0};
        sint4 b[] = {2,1,0,3};
        FORU (i, 4) {
          Vec3<real4> &v = vert[a[i]][b[i]];
          v.y = height_at(v.x, v.z);
        }
      }
    }
  }

  bool all_flat = (flattened[0] && flattened[1] && flattened[2] && flattened[3]);
 
  // offset for tile position
  FORS (i, 4) {
    FORS (j, 4) {
      Vec3<real4> &v = vert[i][j];
      v.x += tx;
      v.z += ty;

      if (!all_flat) {
        bool flat = false;
        if (flattened[i] && flattened[j]) {
          if (i == 0 && j != 3) flat = true;
          if (i == 1 && j != 2) flat = true;
          if (i == 2 && j != 1) flat = true;
          if (i == 3 && j != 0) flat = true;
        }
        if (!flat) v.y += get_noise_at(v.x, v.z);
      }
    }
  }
 
   FORS (i, 4) {
    FORS (j, 4) {
      Vec3<real4> &v = vert[i][j];

      // calculate normal (mostly accurate)

      typedef void (*move_func)(sint4&, sint4&, sint4&, sint4&);
      move_func func[4] = {w_point, e_point, n_point, s_point};
      move_func pfunc[4][6] = {{ne_point, nw_point, se_point, sw_point, n_point, s_point},
                               {ne_point, nw_point, se_point, sw_point, n_point, s_point},
                               {ne_point, nw_point, se_point, sw_point, w_point, e_point},
                               {ne_point, nw_point, se_point, sw_point, w_point, e_point}};

      Vec3<real4> pt[4]; // position of vertices one move away in each compass direction
      sint4 i2, j2, x2, y2;
      
      FORU (q, 4) {
        Vec3<real4> &u = pt[q];
        {
          i2 = i;
          j2 = j;
          x2 = tx;
          y2 = ty;
          do {
            (*func[q])(i2, j2, x2, y2);
            GfxTerrainTile *t = ter->tile(x2, y2);
            if (t == 0) {
              u = v;
              break;
            } else {
              u = t->vert[i2][j2];
            }
          } while (u.x == v.x && u.z == v.z);
        }
        {
          Vec3<real4> s;
          bool second = false;
          FORU (r, 6) {
            i2 = i;
            j2 = j;
            x2 = tx;
            y2 = ty;
            (*pfunc[q][r])(i2, j2, x2, y2);
            GfxTerrainTile *t = ter->tile(x2, y2);
            if (t) {
              Vec3<real4> &w = t->vert[i2][j2];
              if (w.x == v.x && w.z == v.z) { // perpendicular point that is coincident
                FOREVER {
                  (*func[q])(i2, j2, x2, y2);
                  t = ter->tile(x2, y2);
                  if (!t) break;
                  s = t->vert[i2][j2];
                  if (s.x != v.x || s.z != v.z) {
                    second = true;
                    break;
                  }
                }
                if (second) break;
              }
            }
          }
          if (second) { // take the average
            u += s;
            u *= 0.5;
          }
        }
      }

      FORU (q, 4) { pt[q].y += get_noise_at(pt[q].x, pt[q].z); }

      norm[i][j] = (pt[2]-pt[3]).cross(pt[0]-pt[1]);
      norm[i][j].normalize();

      maxy = max(maxy, v.y);
      miny = min(miny, v.y);
    }
  }

  aabb_o.y = miny;
  aabb_d.y = maxy-miny;

  compute_tc();
}

//-------------------------------------------------------------------

void GfxTerrainTile::initblack(sint4 index, GfxTerrain *t)
{
  if (initialized) return;
  ter = t;
  this->index = index;

  sint4 x = index % t->getWidth();
  sint4 y = index / t->getWidth();

  tx = x;
  ty = y;

  GfxTerrainTile *gtt;

  gtt = ter->tile(x-1, y);   if (gtt && gtt->initialized) { h[0] = gtt->h[1]; h[2] = gtt->h[3]; }
  gtt = ter->tile(x+1, y);   if (gtt && gtt->initialized) { h[1] = gtt->h[0]; h[3] = gtt->h[2]; }
  gtt = ter->tile(x, y-1);   if (gtt && gtt->initialized) { h[0] = gtt->h[2]; h[1] = gtt->h[3]; }
  gtt = ter->tile(x, y+1);   if (gtt && gtt->initialized) { h[2] = gtt->h[0]; h[3] = gtt->h[1]; }
  gtt = ter->tile(x-1, y-1); if (gtt && gtt->initialized) { h[0] = gtt->h[3]; }
  gtt = ter->tile(x+1, y+1); if (gtt && gtt->initialized) { h[3] = gtt->h[0]; }
  gtt = ter->tile(x+1, y-1); if (gtt && gtt->initialized) { h[1] = gtt->h[2]; }
  gtt = ter->tile(x-1, y+1); if (gtt && gtt->initialized) { h[2] = gtt->h[1]; }

  has_cliffs = false;

  hf = 1;
  vf = 1;
  trans = 0;
  tt1[0] = 0;
  tt1[1] = sz;

  aabb_o.x = x;
  aabb_o.z = y;
  aabb_d.x = 1;
  aabb_d.z = 1;

  initverts();

//  initialized = true;
}

//-------------------------------------------------------------------

WaterTile* GfxTerrainTile::new_WaterTile()
{
  static real4 depth = 0.25;
  Vec3<real4> watervert[4];
  
  watervert[0] = Vec3<real4>(tx+1, h[1]-depth, ty);
  watervert[1] = Vec3<real4>(tx,   h[0]-depth, ty);
  watervert[2] = Vec3<real4>(tx+1, h[3]-depth, ty+1);
  watervert[3] = Vec3<real4>(tx,   h[2]-depth, ty+1);

  //    FORU(i, 4) watervert[i].y += get_noise_at(watervert[i].x, watervert[i].z);

  return new WaterTile(watervert);
}

//-------------------------------------------------------------------

void GfxTerrainTile::init(sint4 index, Game *game, GfxTerrain *t)
{
  set_gfxm(t->gfxm);
  ter = t;
  this->index = index;

  const real4 inv_h = 1.0/GameConst::HEIGHT_MULT;
  sint4 x = game->get_map().ind2x(index);
  sint4 y = game->get_map().ind2y(index);

  tx = x;
  ty = y;

  const GameTile &tile = game->get_map()(index);

  Random &rng = gfxm->rand;

  sint4 htemp[4]; // get_topo takes integer parameters
  tile.get_topo(htemp[0], htemp[1], htemp[3], htemp[2], split);
  FORS (i, 4) h[i] = inv_h*htemp[i];

  tile.get_type(typeW, typeE);
  GeometryType tW = types[typeW];
  GeometryType tE = types[typeE];

  has_cliffs = (tW == CLIFF || tE == CLIFF);


  // pick texture tiles
  hf = 1;
  vf = 1;
  trans = 0;
  if (tW == WATER && tE != WATER) {

    tt1[0] = sz*2;
    tt1[1] = sz*5;
    if (split == Tile::BT_SPLIT) vf = -1;
    
  } else if (tE == WATER && tW != WATER) {
    tt1[0] = sz*2;
    tt1[1] = sz*5;
    hf = -1;
    if (split==Tile::TB_SPLIT) vf = -1;
    
  } else if (has_water()) {

    tt1[0] = sz * (3+(rng.rand_uint4() % 5));
    tt1[1] = sz*5;
    
  } else {

    tt1[0] = sz * (rng.rand_uint4() % 8);
    tt1[1] = sz * min(int(rng.rand_uint4() % 5), 1);
    hf = (2*(float)(rng.rand_uint4() % 2))-1;
    vf = (2*(float)(rng.rand_uint4() % 2))-1;
    trans=(rng.rand_uint4() % 2);
  }
  
  tt1[0] += sz*(hf<0);
  tt1[1] += sz*(vf<0);

  tt2[0] = sz * (rng.rand_uint4() % 8);
  tt2[1] = 6*sz;


  aabb_o.x = x;
  aabb_o.z = y;
  aabb_d.x = 1;
  aabb_d.z = 1;

  initverts();

  initialized = true;
}

//===================================================================

void GfxTerrainTile::flatten(sint4 f)
{
  flattened[0] += f;
  flattened[1] += f;
  flattened[2] += f;
  flattened[3] += f;
  
  //  if (initialized)
  {
    initverts();

    // align surrounding tiles
    int i = tx;
    int j = ty;

    GfxTerrainTile *t;

    if (i > 0) {
      t = ter->tile(i-1, j); // W tile
      if (t) {
        t->flattened[1] += f;
        t->flattened[3] += f;
        t->initverts();
      }

      t = ter->tile(i-1, j-1); // NW tile
      if (t) {
        t->flattened[3] += f;
        t->initverts();
      }

      t = ter->tile(i-1, j+1); // SW tile
      if (t) {
        t->flattened[1] += f;
        t->initverts();
      }
    }

    t = ter->tile(i, j-1); // N tile
    if (t) {
      t->flattened[2] += f;
      t->flattened[3] += f;
      t->initverts();
    }

    t = ter->tile(i, j+1); // S tile
    if (t) {
      t->flattened[0] += f;
      t->flattened[1] += f;
      t->initverts();
    }

    if (i+1 < ter->getWidth()) {
      t = ter->tile(i+1, j); // E tile
      if (t) {
        t->flattened[0] += f;
        t->flattened[2] += f;
        t->initverts();
      }

      t = ter->tile(i+1, j-1); // NE tile
      if (t) {
        t->flattened[2] += f;
        t->initverts();
      }

      t = ter->tile(i+1, j+1); // SE tile
      if (t) {
        t->flattened[0] += f;
        t->initverts();
      }
    }
  }
}

//===================================================================

void GfxTerrainTile::setTexture(real4 row, real4 col, sint4 ncol,
                                float hflip, float vflip, bool transpose)
{
  tt1[0] = sz * (col+(gfxm->rand.rand_uint4() % ncol));
  tt1[1] = row*sz;
  trans = transpose;
  hf = hflip;
  vf = vflip;
  tt1[0] += sz*(hf<0);
  tt1[1] += sz*(vf<0);
}

//===================================================================

void GfxTerrainTile::NW_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx-1, ty-1);
  if (!t || !t->initialized || !initialized) return;
  h[0] = t->h[3];
  initverts();
}

//-------------------------------------------------------------------

void GfxTerrainTile::N_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx, ty-1);
  if (!t || !t->initialized || !initialized) return;
  h[0] = t->h[2];
  h[1] = t->h[3];
  initverts();
}

//-------------------------------------------------------------------

void GfxTerrainTile::NE_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx+1, ty-1);
  if (!t || !t->initialized || !initialized) return;
  h[1] = t->h[2];
  initverts();
}

//-------------------------------------------------------------------

void GfxTerrainTile::E_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx+1, ty);
  if (!t || !t->initialized || !initialized) return;
  h[1] = t->h[0];
  h[3] = t->h[2];
  initverts();
}

//-------------------------------------------------------------------

void GfxTerrainTile::SE_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx+1, ty+1);
  if (!t || !t->initialized || !initialized) return;
  h[3] = t->h[0];
  initverts();
}

//-------------------------------------------------------------------

void GfxTerrainTile::S_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx, ty+1);
  if (!t || !t->initialized || !initialized) return;
  h[2] = t->h[0];
  h[3] = t->h[1];
  initverts();
}

//-------------------------------------------------------------------

void GfxTerrainTile::SW_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx-1, ty+1);
  if (!t || !t->initialized || !initialized) return;
  h[2] = t->h[1];
  initverts();
}

//-------------------------------------------------------------------

void GfxTerrainTile::W_neighbor()
{
  GfxTerrainTile *t = ter->tile(tx-1, ty);
  if (!t || !t->initialized || !initialized) return;
  h[0] = t->h[1];
  h[2] = t->h[3];
  initverts();
}

//===================================================================
bool GfxTerrainTile::N_neighbor_exists()
{
  GfxTerrainTile *t = ter->tile(tx, ty-1);
  if (!t || !t->initialized || !initialized) 
    return false;
  return true;
}

//-------------------------------------------------------------------

bool GfxTerrainTile::E_neighbor_exists()
{
  GfxTerrainTile *t = ter->tile(tx+1, ty);
  if (!t || !t->initialized || !initialized) 
    return false;
  return true;
}

//-------------------------------------------------------------------

bool GfxTerrainTile::S_neighbor_exists()
{
  GfxTerrainTile *t = ter->tile(tx, ty+1);
  if (!t || !t->initialized || !initialized) 
    return false;
  return true;
}

//-------------------------------------------------------------------

bool GfxTerrainTile::W_neighbor_exists()
{
  GfxTerrainTile *t = ter->tile(tx-1, ty);
  if (!t || !t->initialized || !initialized) 
    return false;
  return true;
}

//===================================================================

#define CNV(i,j)\
glColor3f(vert[i][j].y/8+0.35, vert[i][j].y/8+0.35, vert[i][j].y/8+0.35);\
glNormal3fv(norm[i][j].ptr());\
glVertex3fv(vert[i][j].ptr());
void GfxTerrainTile::draw()
{
  //  draw_shader(); return;

  if (has_cliffs) {

    // non cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2fv(texcoord[0][0]); CNV(0,0);
    glTexCoord2fv(texcoord[0][2]); CNV(0,2);
    glTexCoord2fv(texcoord[0][1]); CNV(0,1);
    glTexCoord2fv(texcoord[0][3]); CNV(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2fv(texcoord[1][0]); CNV(1,0);
    glTexCoord2fv(texcoord[1][2]); CNV(1,2);
    glTexCoord2fv(texcoord[1][1]); CNV(1,1);
    glTexCoord2fv(texcoord[1][3]); CNV(1,3);
    glEnd(); 

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2fv(texcoord[2][0]); CNV(2,0);
    glTexCoord2fv(texcoord[2][2]); CNV(2,2);
    glTexCoord2fv(texcoord[2][1]); CNV(2,1);
    glTexCoord2fv(texcoord[2][3]); CNV(2,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2fv(texcoord[3][0]); CNV(3,0);
    glTexCoord2fv(texcoord[3][2]); CNV(3,2);
    glTexCoord2fv(texcoord[3][1]); CNV(3,1);
    glTexCoord2fv(texcoord[3][3]); CNV(3,3);
    glEnd();

    // cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(tt2[0]+sz*vert[0][1].y, tt2[1]+sz-eps); CNV(0,1);
    glTexCoord2f(tt2[0]+sz*vert[0][3].y, tt2[1]+sz/2);   CNV(0,3);
    glTexCoord2f(tt2[0]+sz*vert[1][0].y, tt2[1]+sz-eps); CNV(1,0);
    glTexCoord2f(tt2[0]+sz*vert[1][2].y, tt2[1]+sz/2);   CNV(1,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(tt2[0]+sz*vert[2][1].y, tt2[1]+sz/2); CNV(2,1);
    glTexCoord2f(tt2[0]+sz*vert[2][3].y, tt2[1]+eps);  CNV(2,3);
    glTexCoord2f(tt2[0]+sz*vert[3][0].y, tt2[1]+sz/2); CNV(3,0);
    glTexCoord2f(tt2[0]+sz*vert[3][2].y, tt2[1]+eps);  CNV(3,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(tt2[0]+sz*vert[2][0].y, tt2[1]+eps);  CNV(2,0);
    glTexCoord2f(tt2[0]+sz*vert[2][1].y, tt2[1]+sz/2); CNV(2,1);
    glTexCoord2f(tt2[0]+sz*vert[0][2].y, tt2[1]+eps);  CNV(0,2);
    glTexCoord2f(tt2[0]+sz*vert[0][3].y, tt2[1]+sz/2); CNV(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(tt2[0]+sz*vert[3][0].y, tt2[1]+sz/2);   CNV(3,0);
    glTexCoord2f(tt2[0]+sz*vert[3][1].y, tt2[1]+sz-eps); CNV(3,1);
    glTexCoord2f(tt2[0]+sz*vert[1][2].y, tt2[1]+sz/2);   CNV(1,2);
    glTexCoord2f(tt2[0]+sz*vert[1][3].y, tt2[1]+sz-eps); CNV(1,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(tt2[0]+sz*vert[1][2].y, tt2[1]+eps);  CNV(1,2);
    glTexCoord2f(tt2[0]+sz*vert[0][3].y, tt2[1]+eps);  CNV(0,3);
    glTexCoord2f(tt2[0]+sz*vert[3][0].y, tt2[1]+sz/4); CNV(3,0);
    glTexCoord2f(tt2[0]+sz*vert[2][1].y, tt2[1]+sz/4); CNV(2,1);
    glEnd();

  } else {

    // no cliffs

    glBegin(GL_TRIANGLE_FAN);
    glTexCoord2fv(texcoord[0][3]); CNV(0,3);//1
    glTexCoord2fv(texcoord[0][0]); CNV(0,0);//2
    glTexCoord2fv(texcoord[0][2]); CNV(0,2);//3
    glTexCoord2fv(texcoord[2][0]); CNV(2,0);//IGNORE
    glTexCoord2fv(texcoord[2][2]); CNV(2,2);//4
    glTexCoord2fv(texcoord[2][3]); CNV(2,3);//5
    glTexCoord2fv(texcoord[3][2]); CNV(3,2);//IGNORE
    glTexCoord2fv(texcoord[3][3]); CNV(3,3);//6
    glTexCoord2fv(texcoord[3][1]); CNV(3,1);//7
    glTexCoord2fv(texcoord[1][3]); CNV(1,3);//IGNORE
    glTexCoord2fv(texcoord[1][1]); CNV(1,1);//8
    glTexCoord2fv(texcoord[1][0]); CNV(1,0);//9
    glTexCoord2fv(texcoord[0][1]); CNV(0,1);//IGNORE
    glTexCoord2fv(texcoord[0][0]); CNV(0,0);//10 (SIMPLY CLOSES THE FAN)
    glEnd();
  }

  if (outline)draw_outline();
}

void GfxTerrainTile::getVerticesForCliffTile(float vertices[432]){
  
  int i, j;

  
  //non-cliff pieces come first;

  i = 0; 
  j = 0;
  vertices[  0]=(vert[i][j].ptr())[0];   vertices[  1]=(vert[i][j].ptr())[1];   vertices[  2]=(vert[i][j].ptr())[2];  //0  position
  vertices[  3]=(norm[i][j].ptr())[0];   vertices[  4]=(norm[i][j].ptr())[1];   vertices[  5]=(norm[i][j].ptr())[2];  //0  normal
  vertices[  6]=  (texcoord[i][j])[0];   vertices[  7]=  (texcoord[i][j])[1];   vertices[  8]=0.0;                    //0  first uvw  map
  vertices[  9]= 0.5f;                   vertices[ 10]= 0.5f;                   vertices[ 11]=0.0;                    //0  second uvw map
  
  i = 0; 
  j = 2;
  vertices[ 12]=(vert[i][j].ptr())[0];   vertices[ 13]=(vert[i][j].ptr())[1];   vertices[ 14]=(vert[i][j].ptr())[2];  //1  position
  vertices[ 15]=(norm[i][j].ptr())[0];   vertices[ 16]=(norm[i][j].ptr())[1];   vertices[ 17]=(norm[i][j].ptr())[2];  //1  normal
  vertices[ 18]=  (texcoord[i][j])[0];   vertices[ 19]=  (texcoord[i][j])[1];   vertices[ 20]=0.0;                    //1  first uvw  map
  vertices[ 21]= 0.0f;                   vertices[ 22]= 0.0f;                   vertices[ 23]=0.0;                    //1  second uvw map
  
  i = 0; 
  j = 1;
  vertices[ 24]=(vert[i][j].ptr())[0];   vertices[ 25]=(vert[i][j].ptr())[1];   vertices[ 26]=(vert[i][j].ptr())[2];  //2  position
  vertices[ 27]=(norm[i][j].ptr())[0];   vertices[ 28]=(norm[i][j].ptr())[1];   vertices[ 29]=(norm[i][j].ptr())[2];  //2  normal
  vertices[ 30]=  (texcoord[i][j])[0];   vertices[ 31]=  (texcoord[i][j])[1];   vertices[ 32]=0.0;                    //2  first uvw  map
  vertices[ 33]= 0.0f;                   vertices[ 34]= 0.5f;                   vertices[ 35]=0.0;                    //2  second uvw map
  
  i = 0; 
  j = 3;
  vertices[ 36]=(vert[i][j].ptr())[0];   vertices[ 37]=(vert[i][j].ptr())[1];   vertices[ 38]=(vert[i][j].ptr())[2];  //3  position
  vertices[ 39]=(norm[i][j].ptr())[0];   vertices[ 40]=(norm[i][j].ptr())[1];   vertices[ 41]=(norm[i][j].ptr())[2];  //3  normal
  vertices[ 42]=  (texcoord[i][j])[0];   vertices[ 43]=  (texcoord[i][j])[1];   vertices[ 44]=0.0;                    //3  first uvw  map
  vertices[ 45]= 0.0f;                   vertices[ 46]= 1.0f;                   vertices[ 47]=0.0;                    //3  second uvw map
  
  i = 1; 
  j = 0;
  vertices[ 48]=(vert[i][j].ptr())[0];   vertices[ 49]=(vert[i][j].ptr())[1];   vertices[ 50]=(vert[i][j].ptr())[2];  //4  position
  vertices[ 51]=(norm[i][j].ptr())[0];   vertices[ 52]=(norm[i][j].ptr())[1];   vertices[ 53]=(norm[i][j].ptr())[2];  //4  normal
  vertices[ 54]=  (texcoord[i][j])[0];   vertices[ 55]=  (texcoord[i][j])[1];   vertices[ 56]=0.0;                    //4  first uvw  map
  vertices[ 57]= 0.5f;                   vertices[ 58]= 1.0f;                   vertices[ 59]=0.0;                    //4  second uvw map
  
  i = 1; 
  j = 2;
  vertices[ 60]=(vert[i][j].ptr())[0];   vertices[ 61]=(vert[i][j].ptr())[1];   vertices[ 62]=(vert[i][j].ptr())[2];  //5  position
  vertices[ 63]=(norm[i][j].ptr())[0];   vertices[ 64]=(norm[i][j].ptr())[1];   vertices[ 65]=(norm[i][j].ptr())[2];  //5  normal
  vertices[ 66]=  (texcoord[i][j])[0];   vertices[ 67]=  (texcoord[i][j])[1];   vertices[ 68]=0.0;                    //5  first uvw  map
  vertices[ 69]= 1.0f;                   vertices[ 70]= 1.0f;                   vertices[ 71]=0.0;                    //5  second uvw map
  
  i = 1; 
  j = 1;
  vertices[ 72]=(vert[i][j].ptr())[0];   vertices[ 73]=(vert[i][j].ptr())[1];   vertices[ 74]=(vert[i][j].ptr())[2];  //6  position
  vertices[ 75]=(norm[i][j].ptr())[0];   vertices[ 76]=(norm[i][j].ptr())[1];   vertices[ 77]=(norm[i][j].ptr())[2];  //6  normal
  vertices[ 78]=  (texcoord[i][j])[0];   vertices[ 79]=  (texcoord[i][j])[1];   vertices[ 80]=0.0;                    //6  first uvw  map
  vertices[ 81]= 1.0f;                   vertices[ 82]= 0.5f;                   vertices[ 83]=0.0;                    //6  second uvw map
  
  i = 1; 
  j = 3;
  vertices[ 84]=(vert[i][j].ptr())[0];   vertices[ 85]=(vert[i][j].ptr())[1];   vertices[ 86]=(vert[i][j].ptr())[2];  //7  position
  vertices[ 87]=(norm[i][j].ptr())[0];   vertices[ 88]=(norm[i][j].ptr())[1];   vertices[ 89]=(norm[i][j].ptr())[2];  //7  normal
  vertices[ 90]=  (texcoord[i][j])[0];   vertices[ 91]=  (texcoord[i][j])[1];   vertices[ 92]=0.0;                    //7  first uvw  map
  vertices[ 93]= 1.0f;                   vertices[ 94]= 0.0f;                   vertices[ 95]=0.0;                    //7  second uvw map
  
  i = 2; 
  j = 0;
  vertices[ 96]=(vert[i][j].ptr())[0];   vertices[ 97]=(vert[i][j].ptr())[1];   vertices[ 98]=(vert[i][j].ptr())[2];  //8  position
  vertices[ 99]=(norm[i][j].ptr())[0];   vertices[100]=(norm[i][j].ptr())[1];   vertices[101]=(norm[i][j].ptr())[2];  //8  normal
  vertices[102]=  (texcoord[i][j])[0];   vertices[103]=  (texcoord[i][j])[1];   vertices[104]=0.0;                    //8  first uvw  map
  vertices[105]= 0.5f;                   vertices[106]= 0.0f;                   vertices[107]=0.0;                    //8  second uvw map
  
  i = 2; 
  j = 2;
  vertices[108]=(vert[i][j].ptr())[0];   vertices[109]=(vert[i][j].ptr())[1];   vertices[110]=(vert[i][j].ptr())[2];  //9  position
  vertices[111]=(norm[i][j].ptr())[0];   vertices[112]=(norm[i][j].ptr())[1];   vertices[113]=(norm[i][j].ptr())[2];  //9  normal
  vertices[114]=  (texcoord[i][j])[0];   vertices[115]=  (texcoord[i][j])[1];   vertices[116]=0.0;                    //9  first uvw  map
  vertices[117]= 0.0f;                   vertices[118]= 0.0f;                   vertices[119]=0.0;                    //9  second uvw map
  
  i = 2; 
  j = 1;
  vertices[120]=(vert[i][j].ptr())[0];   vertices[121]=(vert[i][j].ptr())[1];   vertices[122]=(vert[i][j].ptr())[2];  //10 position
  vertices[123]=(norm[i][j].ptr())[0];   vertices[124]=(norm[i][j].ptr())[1];   vertices[125]=(norm[i][j].ptr())[2];  //10 normal
  vertices[126]=  (texcoord[i][j])[0];   vertices[127]=  (texcoord[i][j])[1];   vertices[128]=0.0;                    //10 first uvw  map
  vertices[129]= 0.0f;                   vertices[130]= 0.5f;                   vertices[131]=0.0;                    //10 second uvw map
  
  i = 2; 
  j = 3;
  vertices[132]=(vert[i][j].ptr())[0];   vertices[133]=(vert[i][j].ptr())[1];   vertices[134]=(vert[i][j].ptr())[2];  //11 position
  vertices[135]=(norm[i][j].ptr())[0];   vertices[136]=(norm[i][j].ptr())[1];   vertices[137]=(norm[i][j].ptr())[2];  //11 normal
  vertices[138]=  (texcoord[i][j])[0];   vertices[139]=  (texcoord[i][j])[1];   vertices[140]=0.0;                    //11 first uvw  map
  vertices[141]= 0.0f;                   vertices[142]= 1.0f;                   vertices[143]=0.0;                    //11 second uvw map
  
  i = 3; 
  j = 0;
  vertices[144]=(vert[i][j].ptr())[0];   vertices[145]=(vert[i][j].ptr())[1];   vertices[146]=(vert[i][j].ptr())[2];  //12 position
  vertices[147]=(norm[i][j].ptr())[0];   vertices[148]=(norm[i][j].ptr())[1];   vertices[149]=(norm[i][j].ptr())[2];  //12 normal
  vertices[150]=  (texcoord[i][j])[0];   vertices[151]=  (texcoord[i][j])[1];   vertices[152]=0.0;                    //12 first uvw  map
  vertices[153]= 0.5f;                   vertices[154]= 1.0f;                   vertices[155]=0.0;                    //12 second uvw map
  
  i = 3; 
  j = 2;
  vertices[156]=(vert[i][j].ptr())[0];   vertices[157]=(vert[i][j].ptr())[1];   vertices[158]=(vert[i][j].ptr())[2];  //13 position
  vertices[159]=(norm[i][j].ptr())[0];   vertices[160]=(norm[i][j].ptr())[1];   vertices[161]=(norm[i][j].ptr())[2];  //13 normal
  vertices[162]=  (texcoord[i][j])[0];   vertices[163]=  (texcoord[i][j])[1];   vertices[164]=0.0;                    //13 first uvw  map
  vertices[165]= 1.0f;                   vertices[166]= 1.0f;                   vertices[167]=0.0;                    //13 second uvw map
  
  i = 3; 
  j = 1;
  vertices[168]=(vert[i][j].ptr())[0];   vertices[169]=(vert[i][j].ptr())[1];   vertices[170]=(vert[i][j].ptr())[2];  //14 position
  vertices[171]=(norm[i][j].ptr())[0];   vertices[172]=(norm[i][j].ptr())[1];   vertices[173]=(norm[i][j].ptr())[2];  //14 normal
  vertices[174]=  (texcoord[i][j])[0];   vertices[175]=  (texcoord[i][j])[1];   vertices[176]=0.0;                    //14 first uvw  map
  vertices[177]= 1.0f;                   vertices[178]= 0.5f;                   vertices[179]=0.0;                    //14 second uvw map
  
  i = 3; 
  j = 3;
  vertices[180]=(vert[i][j].ptr())[0];   vertices[181]=(vert[i][j].ptr())[1];   vertices[182]=(vert[i][j].ptr())[2];  //15 position
  vertices[183]=(norm[i][j].ptr())[0];   vertices[184]=(norm[i][j].ptr())[1];   vertices[185]=(norm[i][j].ptr())[2];  //15 normal
  vertices[186]=  (texcoord[i][j])[0];   vertices[187]=  (texcoord[i][j])[1];   vertices[188]=0.0;                    //15 first uvw  map
  vertices[189]= 1.0f;                   vertices[190]= 0.0f;                   vertices[191]=0.0;                    //15 second uvw map
  
  // lastly, all our precious rocky-cliff pieces come.
  
  i = 0; 
  j = 1;
  vertices[192]=(vert[i][j].ptr())[0];   vertices[193]=(vert[i][j].ptr())[1];   vertices[194]=(vert[i][j].ptr())[2];  //16 position
  vertices[195]=(norm[i][j].ptr())[0];   vertices[196]=(norm[i][j].ptr())[1];   vertices[197]=(norm[i][j].ptr())[2];  //16 normal
  vertices[198]=tt2[0]+sz*vert[i][j].y;  vertices[199]= tt2[1]+sz-eps;          vertices[200]=0.0;                    //16 first uvw  map
  vertices[201]= 1.0f;                   vertices[202]= 0.0f;                   vertices[203]=0.0;                    //16 second uvw map
  
  i = 0; 
  j = 3;
  vertices[204]=(vert[i][j].ptr())[0];   vertices[205]=(vert[i][j].ptr())[1];   vertices[206]=(vert[i][j].ptr())[2];  //17 position
  vertices[207]=(norm[i][j].ptr())[0];   vertices[208]=(norm[i][j].ptr())[1];   vertices[209]=(norm[i][j].ptr())[2];  //17 normal
  vertices[210]=tt2[0]+sz*vert[i][j].y;  vertices[211]= tt2[1]+sz/2;            vertices[212]=0.0;                    //17 first uvw  map
  vertices[213]= 1.0f;                   vertices[214]= 0.0f;                   vertices[215]=0.0;                    //17 second uvw map
  
  i = 1; 
  j = 0;
  vertices[216]=(vert[i][j].ptr())[0];   vertices[217]=(vert[i][j].ptr())[1];   vertices[218]=(vert[i][j].ptr())[2];  //18 position
  vertices[219]=(norm[i][j].ptr())[0];   vertices[220]=(norm[i][j].ptr())[1];   vertices[221]=(norm[i][j].ptr())[2];  //18 normal
  vertices[222]=tt2[0]+sz*vert[i][j].y;  vertices[223]= tt2[1]+sz-eps;          vertices[224]=0.0;                    //18 first uvw  map
  vertices[225]= 1.0f;                   vertices[226]= 0.0f;                   vertices[227]=0.0;                    //18 second uvw map
  
  i = 1; 
  j = 2;
  vertices[228]=(vert[i][j].ptr())[0];   vertices[229]=(vert[i][j].ptr())[1];   vertices[230]=(vert[i][j].ptr())[2];  //19 position
  vertices[231]=(norm[i][j].ptr())[0];   vertices[232]=(norm[i][j].ptr())[1];   vertices[233]=(norm[i][j].ptr())[2];  //19 normal
  vertices[234]=tt2[0]+sz*vert[i][j].y;  vertices[235]= tt2[1]+sz/2;            vertices[236]=0.0;                    //19 first uvw  map
  vertices[237]= 1.0f;                   vertices[238]= 0.0f;                   vertices[239]=0.0;                    //19 second uvw map
  
  

  i = 2; 
  j = 1;
  vertices[240]=(vert[i][j].ptr())[0];   vertices[241]=(vert[i][j].ptr())[1];   vertices[242]=(vert[i][j].ptr())[2];  //20 position
  vertices[243]=(norm[i][j].ptr())[0];   vertices[244]=(norm[i][j].ptr())[1];   vertices[245]=(norm[i][j].ptr())[2];  //20 normal
  vertices[246]=tt2[0]+sz*vert[i][j].y;  vertices[247]= tt2[1]+sz/2;            vertices[248]=0.0;                    //20 first uvw  map
  vertices[249]= 1.0f;                   vertices[250]= 0.0f;                   vertices[251]=0.0;                    //20 second uvw map
  
  i = 2; 
  j = 3;
  vertices[252]=(vert[i][j].ptr())[0];   vertices[253]=(vert[i][j].ptr())[1];   vertices[254]=(vert[i][j].ptr())[2];  //21 position
  vertices[255]=(norm[i][j].ptr())[0];   vertices[256]=(norm[i][j].ptr())[1];   vertices[257]=(norm[i][j].ptr())[2];  //21 normal
  vertices[258]=tt2[0]+sz*vert[i][j].y;  vertices[259]= tt2[1]+eps;             vertices[260]=0.0;                    //21 first uvw  map
  vertices[261]= 1.0f;                   vertices[262]= 0.0f;                   vertices[263]=0.0;                    //21 second uvw map
  
  i = 3; 
  j = 0;
  vertices[264]=(vert[i][j].ptr())[0];   vertices[265]=(vert[i][j].ptr())[1];   vertices[266]=(vert[i][j].ptr())[2];  //22 position
  vertices[267]=(norm[i][j].ptr())[0];   vertices[268]=(norm[i][j].ptr())[1];   vertices[269]=(norm[i][j].ptr())[2];  //22 normal
  vertices[270]=tt2[0]+sz*vert[i][j].y;  vertices[271]= tt2[1]+sz/2;            vertices[272]=0.0;                    //22 first uvw  map
  vertices[273]= 1.0f;                   vertices[274]= 0.0f;                   vertices[275]=0.0;                    //22 second uvw map
  
  i = 3; 
  j = 2;
  vertices[276]=(vert[i][j].ptr())[0];   vertices[277]=(vert[i][j].ptr())[1];   vertices[278]=(vert[i][j].ptr())[2];  //23 position
  vertices[279]=(norm[i][j].ptr())[0];   vertices[280]=(norm[i][j].ptr())[1];   vertices[281]=(norm[i][j].ptr())[2];  //23 normal
  vertices[282]=tt2[0]+sz*vert[i][j].y;  vertices[283]= tt2[1]+eps;             vertices[284]=0.0;                    //23 first uvw  map
  vertices[285]= 1.0f;                   vertices[286]= 0.0f;                   vertices[287]=0.0;                    //23 second uvw map
  
  

  i = 2; 
  j = 0;
  vertices[288]=(vert[i][j].ptr())[0];   vertices[289]=(vert[i][j].ptr())[1];   vertices[290]=(vert[i][j].ptr())[2];  //24 position
  vertices[291]=(norm[i][j].ptr())[0];   vertices[292]=(norm[i][j].ptr())[1];   vertices[293]=(norm[i][j].ptr())[2];  //24 normal
  vertices[294]=tt2[0]+sz*vert[i][j].y;  vertices[295]= tt2[1]+eps;             vertices[296]=0.0;                    //24 first uvw  map
  vertices[297]= 1.0f;                   vertices[298]= 0.0f;                   vertices[299]=0.0;                    //24 second uvw map
  
  i = 2; 
  j = 1;
  vertices[300]=(vert[i][j].ptr())[0];   vertices[301]=(vert[i][j].ptr())[1];   vertices[302]=(vert[i][j].ptr())[2];  //25 position
  vertices[303]=(norm[i][j].ptr())[0];   vertices[304]=(norm[i][j].ptr())[1];   vertices[305]=(norm[i][j].ptr())[2];  //25 normal
  vertices[306]=tt2[0]+sz*vert[i][j].y;  vertices[307]= tt2[1]+sz/2;            vertices[308]=0.0;                    //25 first uvw  map
  vertices[309]= 1.0f;                   vertices[310]= 0.0f;                   vertices[311]=0.0;                    //25 second uvw map
  
  i = 0; 
  j = 2;
  vertices[312]=(vert[i][j].ptr())[0];   vertices[313]=(vert[i][j].ptr())[1];   vertices[314]=(vert[i][j].ptr())[2];  //26 position
  vertices[315]=(norm[i][j].ptr())[0];   vertices[316]=(norm[i][j].ptr())[1];   vertices[317]=(norm[i][j].ptr())[2];  //26 normal
  vertices[318]=tt2[0]+sz*vert[i][j].y;  vertices[319]= tt2[1]+eps;             vertices[320]=0.0;                    //26 first uvw  map
  vertices[321]= 1.0f;                   vertices[322]= 0.0f;                   vertices[323]=0.0;                    //26 second uvw map
  
  i = 0; 
  j = 3;
  vertices[324]=(vert[i][j].ptr())[0];   vertices[325]=(vert[i][j].ptr())[1];   vertices[326]=(vert[i][j].ptr())[2];  //27 position
  vertices[327]=(norm[i][j].ptr())[0];   vertices[328]=(norm[i][j].ptr())[1];   vertices[329]=(norm[i][j].ptr())[2];  //27 normal
  vertices[330]=tt2[0]+sz*vert[i][j].y;  vertices[331]= tt2[1]+sz/2;            vertices[332]=0.0;                    //27 first uvw  map
  vertices[333]= 1.0f;                   vertices[334]= 0.0f;                   vertices[335]=0.0;                    //27 second uvw map
  
  

  i = 3; 
  j = 0;
  vertices[336]=(vert[i][j].ptr())[0];   vertices[337]=(vert[i][j].ptr())[1];   vertices[338]=(vert[i][j].ptr())[2];  //28 position
  vertices[339]=(norm[i][j].ptr())[0];   vertices[340]=(norm[i][j].ptr())[1];   vertices[341]=(norm[i][j].ptr())[2];  //28 normal
  vertices[342]=tt2[0]+sz*vert[i][j].y;  vertices[343]= tt2[1]+sz/2;            vertices[344]=0.0;                    //28 first uvw  map
  vertices[345]= 1.0f;                   vertices[346]= 0.0f;                   vertices[347]=0.0;                    //28 second uvw map
  
  i = 3; 
  j = 1;
  vertices[348]=(vert[i][j].ptr())[0];   vertices[349]=(vert[i][j].ptr())[1];   vertices[350]=(vert[i][j].ptr())[2];  //29 position
  vertices[351]=(norm[i][j].ptr())[0];   vertices[352]=(norm[i][j].ptr())[1];   vertices[353]=(norm[i][j].ptr())[2];  //29 normal
  vertices[354]=tt2[0]+sz*vert[i][j].y;  vertices[355]= tt2[1]+sz-eps;          vertices[356]=0.0;                    //29 first uvw  map
  vertices[357]= 1.0f;                   vertices[358]= 0.0f;                   vertices[359]=0.0;                    //29 second uvw map
  
  i = 1; 
  j = 2;
  vertices[360]=(vert[i][j].ptr())[0];   vertices[361]=(vert[i][j].ptr())[1];   vertices[362]=(vert[i][j].ptr())[2];  //30 position
  vertices[363]=(norm[i][j].ptr())[0];   vertices[364]=(norm[i][j].ptr())[1];   vertices[365]=(norm[i][j].ptr())[2];  //30 normal
  vertices[366]=tt2[0]+sz*vert[i][j].y;  vertices[367]= tt2[1]+sz/2;            vertices[368]=0.0;                    //30 first uvw  map
  vertices[369]= 1.0f;                   vertices[370]= 0.0f;                   vertices[371]=0.0;                    //30 second uvw map
  
  i = 1; 
  j = 3;
  vertices[372]=(vert[i][j].ptr())[0];   vertices[373]=(vert[i][j].ptr())[1];   vertices[374]=(vert[i][j].ptr())[2];  //31 position
  vertices[375]=(norm[i][j].ptr())[0];   vertices[376]=(norm[i][j].ptr())[1];   vertices[377]=(norm[i][j].ptr())[2];  //31 normal
  vertices[378]=tt2[0]+sz*vert[i][j].y;  vertices[379]= tt2[1]+sz-eps;          vertices[380]=0.0;                    //31 first uvw  map
  vertices[381]= 1.0f;                   vertices[382]= 0.0f;                   vertices[383]=0.0;                    //31 second uvw map
  
  

  i = 1; 
  j = 2;
  vertices[384]=(vert[i][j].ptr())[0];   vertices[385]=(vert[i][j].ptr())[1];   vertices[386]=(vert[i][j].ptr())[2];  //32 position
  vertices[387]=(norm[i][j].ptr())[0];   vertices[388]=(norm[i][j].ptr())[1];   vertices[389]=(norm[i][j].ptr())[2];  //32 normal
  vertices[390]=tt2[0]+sz*vert[i][j].y;  vertices[391]= tt2[1]+eps;             vertices[392]=0.0;                    //32 first uvw  map
  vertices[393]= 1.0f;                   vertices[394]= 0.0f;                   vertices[395]=0.0;                    //32 second uvw map
  
  i = 0; 
  j = 3;
  vertices[396]=(vert[i][j].ptr())[0];   vertices[397]=(vert[i][j].ptr())[1];   vertices[398]=(vert[i][j].ptr())[2];  //33 position
  vertices[399]=(norm[i][j].ptr())[0];   vertices[400]=(norm[i][j].ptr())[1];   vertices[401]=(norm[i][j].ptr())[2];  //33 normal
  vertices[402]=tt2[0]+sz*vert[i][j].y;  vertices[403]= tt2[1]+eps;             vertices[404]=0.0;                    //33 first uvw  map
  vertices[405]= 1.0f;                   vertices[406]= 0.0f;                   vertices[407]=0.0;                    //33 second uvw map
  
  i = 3; 
  j = 0;
  vertices[408]=(vert[i][j].ptr())[0];   vertices[409]=(vert[i][j].ptr())[1];   vertices[410]=(vert[i][j].ptr())[2];  //34 position
  vertices[411]=(norm[i][j].ptr())[0];   vertices[412]=(norm[i][j].ptr())[1];   vertices[413]=(norm[i][j].ptr())[2];  //34 normal
  vertices[414]=tt2[0]+sz*vert[i][j].y;  vertices[415]= tt2[1]+sz/4;            vertices[416]=0.0;                    //34 first uvw  map
  vertices[417]= 1.0f;                   vertices[418]= 0.0f;                   vertices[419]=0.0;                    //34 second uvw map
  
  i = 2; 
  j = 1;
  vertices[420]=(vert[i][j].ptr())[0];   vertices[421]=(vert[i][j].ptr())[1];   vertices[422]=(vert[i][j].ptr())[2];  //35 position
  vertices[423]=(norm[i][j].ptr())[0];   vertices[424]=(norm[i][j].ptr())[1];   vertices[425]=(norm[i][j].ptr())[2];  //35 normal
  vertices[426]=tt2[0]+sz*vert[i][j].y;  vertices[427]= tt2[1]+sz/4;            vertices[428]=0.0;                    //35 first uvw  map
  vertices[429]= 1.0f;                   vertices[430]= 0.0f;                   vertices[431]=0.0;                    //35 second uvw map
    
  return;
}

void GfxTerrainTile::getVerticesForNonCliffTile(float vertices[108]){
  int i, j;

  //	1 -- 8 -- 7
  //	| \  |  / |
  //	|  \ | /  |
  //	2 -- 0 -- 6
  //	|  / | \  |
  //	| /  |  \ |
  //	3 -- 4 -- 5

  i = 0; 
  j = 3;
  vertices[  0]=(vert[i][j].ptr())[0];   vertices[  1]=(vert[i][j].ptr())[1];   vertices[  2]=(vert[i][j].ptr())[2];  //0 position
  vertices[  3]=(norm[i][j].ptr())[0];   vertices[  4]=(norm[i][j].ptr())[1];   vertices[  5]=(norm[i][j].ptr())[2];  //0 normal
  vertices[  6]=  (texcoord[i][j])[0];   vertices[  7]=  (texcoord[i][j])[1];   vertices[  8]=0.0;                    //0 first uvw  map
  vertices[  9]= 0.5f;                   vertices[ 10]= 0.5f;                   vertices[ 11]=0.0;                    //0 second uvw map
  
  i = 0; 
  j = 0;
  vertices[ 12]=(vert[i][j].ptr())[0];   vertices[ 13]=(vert[i][j].ptr())[1];   vertices[ 14]=(vert[i][j].ptr())[2];  //1 position
  vertices[ 15]=(norm[i][j].ptr())[0];   vertices[ 16]=(norm[i][j].ptr())[1];   vertices[ 17]=(norm[i][j].ptr())[2];  //1 normal
  vertices[ 18]=  (texcoord[i][j])[0];   vertices[ 19]=  (texcoord[i][j])[1];   vertices[ 20]=0.0;                    //1 first uvw  map
  vertices[ 21]= 0.0f;                   vertices[ 22]= 0.0f;                   vertices[ 23]=0.0;                    //1 second uvw map
  
  i = 0; 
  j = 2;
  vertices[ 24]=(vert[i][j].ptr())[0];   vertices[ 25]=(vert[i][j].ptr())[1];   vertices[ 26]=(vert[i][j].ptr())[2];  //2 position
  vertices[ 27]=(norm[i][j].ptr())[0];   vertices[ 28]=(norm[i][j].ptr())[1];   vertices[ 29]=(norm[i][j].ptr())[2];  //2 normal
  vertices[ 30]=  (texcoord[i][j])[0];   vertices[ 31]=  (texcoord[i][j])[1];   vertices[ 32]=0.0;                    //2 first uvw  map
  vertices[ 33]= 0.0f;                   vertices[ 34]= 0.5f;                   vertices[ 35]=0.0;                    //2 second uvw map
  
  i = 2; 
  j = 2;
  vertices[ 36]=(vert[i][j].ptr())[0];   vertices[ 37]=(vert[i][j].ptr())[1];   vertices[ 38]=(vert[i][j].ptr())[2];  //3 position
  vertices[ 39]=(norm[i][j].ptr())[0];   vertices[ 40]=(norm[i][j].ptr())[1];   vertices[ 41]=(norm[i][j].ptr())[2];  //3 normal
  vertices[ 42]=  (texcoord[i][j])[0];   vertices[ 43]=  (texcoord[i][j])[1];   vertices[ 44]=0.0;                    //3 first uvw  map
  vertices[ 45]= 0.0f;                   vertices[ 46]= 1.0f;                   vertices[ 47]=0.0;                    //3 second uvw map
  
  i = 2; 
  j = 3;
  vertices[ 48]=(vert[i][j].ptr())[0];   vertices[ 49]=(vert[i][j].ptr())[1];   vertices[ 50]=(vert[i][j].ptr())[2];  //4 position
  vertices[ 51]=(norm[i][j].ptr())[0];   vertices[ 52]=(norm[i][j].ptr())[1];   vertices[ 53]=(norm[i][j].ptr())[2];  //4 normal
  vertices[ 54]=  (texcoord[i][j])[0];   vertices[ 55]=  (texcoord[i][j])[1];   vertices[ 56]=0.0;                    //4 first uvw  map
  vertices[ 57]= 0.5f;                   vertices[ 58]= 1.0f;                   vertices[ 59]=0.0;                    //4 second uvw map
  
  i = 3; 
  j = 3;
  vertices[ 60]=(vert[i][j].ptr())[0];   vertices[ 61]=(vert[i][j].ptr())[1];   vertices[ 62]=(vert[i][j].ptr())[2];  //5 position
  vertices[ 63]=(norm[i][j].ptr())[0];   vertices[ 64]=(norm[i][j].ptr())[1];   vertices[ 65]=(norm[i][j].ptr())[2];  //5 normal
  vertices[ 66]=  (texcoord[i][j])[0];   vertices[ 67]=  (texcoord[i][j])[1];   vertices[ 68]=0.0;                    //5 first uvw  map
  vertices[ 69]= 1.0f;                   vertices[ 70]= 1.0f;                   vertices[ 71]=0.0;                    //5 second uvw map
  
  i = 3; 
  j = 1;
  vertices[ 72]=(vert[i][j].ptr())[0];   vertices[ 73]=(vert[i][j].ptr())[1];   vertices[ 74]=(vert[i][j].ptr())[2];  //6 position
  vertices[ 75]=(norm[i][j].ptr())[0];   vertices[ 76]=(norm[i][j].ptr())[1];   vertices[ 77]=(norm[i][j].ptr())[2];  //6 normal
  vertices[ 78]=  (texcoord[i][j])[0];   vertices[ 79]=  (texcoord[i][j])[1];   vertices[ 80]=0.0;                    //6 first uvw  map
  vertices[ 81]= 1.0f;                   vertices[ 82]= 0.5f;                   vertices[ 83]=0.0;                    //6 second uvw map
  
  i = 1; 
  j = 1;
  vertices[ 84]=(vert[i][j].ptr())[0];   vertices[ 85]=(vert[i][j].ptr())[1];   vertices[ 86]=(vert[i][j].ptr())[2];  //7 position
  vertices[ 87]=(norm[i][j].ptr())[0];   vertices[ 88]=(norm[i][j].ptr())[1];   vertices[ 89]=(norm[i][j].ptr())[2];  //7 normal
  vertices[ 90]=  (texcoord[i][j])[0];   vertices[ 91]=  (texcoord[i][j])[1];   vertices[ 92]=0.0;                    //7 first uvw  map
  vertices[ 93]= 1.0f;                   vertices[ 94]= 0.0f;                   vertices[ 95]=0.0;                    //7 second uvw map
  
  i = 1; 
  j = 0;
  vertices[ 96]=(vert[i][j].ptr())[0];   vertices[ 97]=(vert[i][j].ptr())[1];   vertices[ 98]=(vert[i][j].ptr())[2];  //8 position
  vertices[ 99]=(norm[i][j].ptr())[0];   vertices[100]=(norm[i][j].ptr())[1];   vertices[101]=(norm[i][j].ptr())[2];  //8 normal
  vertices[102]=  (texcoord[i][j])[0];   vertices[103]=  (texcoord[i][j])[1];   vertices[104]=0.0;                    //8 first uvw  map
  vertices[105]= 0.5f;                   vertices[106]= 0.0f;                   vertices[107]=0.0;                    //8 second uvw map
  
  /*	
  vertices[ 12]= .5;   vertices[ 13]= N ;   vertices[ 14]=0.0;  //1 position
  vertices[ 15]=0.0;   vertices[ 16]=1.0;   vertices[ 17]=0.0;  //1 normal
  vertices[ 18]=UN_1;  vertices[ 19]=VN_1;  vertices[ 20]=0.0;  //1 first uvw  map
  vertices[ 21]=UN_2;  vertices[ 22]=VN_2;  vertices[ 23]=0.0;  //1 second uvw map
		
  vertices[ 24]=1.0;   vertices[ 25]= NE;   vertices[ 26]=0.0;  //2 position
  vertices[ 27]=0.0;   vertices[ 28]=1.0;   vertices[ 29]=0.0;  //2 normal
  vertices[ 30]=UNE_1; vertices[ 31]=VNE_1; vertices[ 32]=0.0;  //2 first uvw  map
  vertices[ 33]=UNE_2; vertices[ 34]=VNE_2; vertices[ 35]=0.0;  //2 second uvw map
		
  vertices[ 36]=0.0;   vertices[ 37]= W ;   vertices[ 38]= .5;  //3 position
  vertices[ 39]=0.0;   vertices[ 40]=1.0;   vertices[ 41]=0.0;  //3 normal
  vertices[ 42]=UW_1;  vertices[ 43]=VW_1;  vertices[ 44]=0.0;  //3 first uvw  map
  vertices[ 45]=UW_2;  vertices[ 46]=VW_2;  vertices[ 47]=0.0;  //3 second uvw map
  
  vertices[ 48]= .5;   vertices[ 49]=MID;   vertices[ 50]= .5;  //4 position
  vertices[ 51]=0.0;   vertices[ 52]=1.0;   vertices[ 53]=0.0;  //4 normal
  vertices[ 54]=UMD_1; vertices[ 55]=VMD_1; vertices[ 56]=0.0;  //4 first uvw  map
  vertices[ 57]=UMD_2; vertices[ 58]=VMD_2; vertices[ 59]=0.0;  //4 second uvw map
		
  vertices[ 60]=1.0;   vertices[ 61]= E ;   vertices[ 62]= .5;  //5 position
  vertices[ 63]=0.0;   vertices[ 64]=1.0;   vertices[ 65]=0.0;  //5 normal
  vertices[ 66]=UE_1;  vertices[ 67]=VE_1;  vertices[ 68]=0.0;  //5 first uvw  map
  vertices[ 69]=UE_2;  vertices[ 70]=VE_2;  vertices[ 71]=0.0;  //5 second uvw map
		
  vertices[ 72]=0.0;   vertices[ 73]= SW;   vertices[ 74]=1.0;  //6 position
  vertices[ 75]=0.0;   vertices[ 76]=1.0;   vertices[ 77]=0.0;  //6 normal
  vertices[ 78]=USW_1; vertices[ 79]=VSW_1; vertices[ 80]=0.0;  //6 first uvw  map
  vertices[ 81]=USW_2; vertices[ 82]=VSW_2; vertices[ 83]=0.0;  //6 second uvw map
		
  vertices[ 84]= .5;   vertices[ 85]= S ;   vertices[ 86]=1.0;  //7 position
  vertices[ 87]=0.0;   vertices[ 88]=1.0;   vertices[ 89]=0.0;  //7 normal
  vertices[ 90]=US_1;  vertices[ 91]=VS_1;  vertices[ 92]=0.0;  //7 first uvw  map
  vertices[ 93]=US_2;  vertices[ 94]=VS_2;  vertices[ 95]=0.0;  //7 second uvw map
		
  vertices[ 96]=1.0;   vertices[ 97]= SE;   vertices[ 98]=1.0;  //8 position
  vertices[ 99]=0.0;   vertices[100]=1.0;   vertices[101]=0.0;  //8 normal
  vertices[102]=USE_1; vertices[103]=VSE_1; vertices[104]=0.0;  //8 first uvw  map
  vertices[105]=USE_2; vertices[106]=VSE_2; vertices[107]=0.0;  //8 second uvw map
  //That should be all */
 
  return;
}

//===================================================================

#define STCNV(i,j)                                              \
  c = vert[i][j].y/8+0.35;                                      \
  glColor3f(c,c,c);                                             \
  glNormal3fv(norm[i][j].ptr());                                \
  glVertex3fv(vert[i][j].ptr());

/*
  if (GfxTerrainTile::tile_shader.begin()) {
    GLint shadow_tex = glGetUniformLocation(GfxTerrainTile::tile_shader.id(), "shadow");
    glUniform1i(shadow_tex, 1);
  }
*/



void GfxTerrainTile::draw_shader()
{
  /*
  glerror("ds1");
  glerror("ds3");
  GLint ntex = glGetUniformLocation(tile_shader.id(), "ntex");
  glerror("ds4");
  cout << "ntex = " << tile_shader.id() << " " << ntex << endl;
  glUniform1i(ntex, 0);
  glerror("ds5");
  */
  
  real4 c;

  if (has_cliffs) {

    // non cliff pieces

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(0,1);
    STCNV(0,0);
    STCNV(0,3);
    STCNV(0,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(1,0);
    STCNV(1,2);
    STCNV(1,1);
    STCNV(1,3);
    glEnd(); 

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(2,0);
    STCNV(2,2);
    STCNV(2,1);
    STCNV(2,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(3,1);
    STCNV(3,0);
    STCNV(3,3);
    STCNV(3,2);
    glEnd();

    // cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    STCNV(0,1);
    STCNV(0,3);
    STCNV(1,0);
    STCNV(1,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(2,1);
    STCNV(2,3);
    STCNV(3,0);
    STCNV(3,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(2,0);
    STCNV(2,1);
    STCNV(0,2);
    STCNV(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(3,0);
    STCNV(3,1);
    STCNV(1,2);
    STCNV(1,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    STCNV(1,2);
    STCNV(0,3);
    STCNV(3,0);
    STCNV(2,1);
    glEnd();

  } else {

    // no cliffs

    glBegin(GL_TRIANGLE_FAN);

    STCNV(0,3);
    STCNV(0,0);
    STCNV(0,2);
    STCNV(2,0);
    STCNV(2,2);
    STCNV(2,3);
    STCNV(3,2);
    STCNV(3,3);
    STCNV(3,1);
    STCNV(1,3);
    STCNV(1,1);
    STCNV(1,0);
    STCNV(0,1);
    STCNV(0,0);
    glEnd();

  }

  if (outline) draw_outline();
}

//-------------------------------------------------------------------

#define CNVf(i,j)\
glNormal3fv(norm[i][j].ptr());\
glVertex3fv(vert[i][j].ptr());

/** Draw the tile while ignoring texture.
    Combine with blending to highlight certain tiles.
*/

void GfxTerrainTile::draw_flat()
{
  if (has_cliffs) {

    // non cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    CNVf(0,0);
    CNVf(0,2);
    CNVf(0,1);
    CNVf(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(1,0);
    CNVf(1,2);
    CNVf(1,1);
    CNVf(1,3);
    glEnd(); 

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(2,0);
    CNVf(2,2);
    CNVf(2,1);
    CNVf(2,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(3,0);
    CNVf(3,2);
    CNVf(3,1);
    CNVf(3,3);
    glEnd();

    // cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    CNVf(0,1);
    CNVf(0,3);
    CNVf(1,0);
    CNVf(1,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(2,1);
    CNVf(2,3);
    CNVf(3,0);
    CNVf(3,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(2,0);
    CNVf(2,1);
    CNVf(0,2);
    CNVf(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(3,0);
    CNVf(3,1);
    CNVf(1,2);
    CNVf(1,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVf(1,2);
    CNVf(0,3);
    CNVf(3,0);
    CNVf(2,1);
    glEnd();
  } else {

    // no cliffs

    glBegin(GL_TRIANGLE_FAN);
    CNVf(0,3);
    CNVf(0,0);
    CNVf(0,2);
    CNVf(2,0);
    CNVf(2,2);
    CNVf(2,3);
    CNVf(3,2);
    CNVf(3,3);
    CNVf(3,1);
    CNVf(1,3);
    CNVf(1,1);
    CNVf(1,0);
    CNVf(0,1);
    CNVf(0,0);
    glEnd();
  }
  if (outline) draw_outline();
}

//-------------------------------------------------------------------

#define CNVp(i,j)\
glColor3fv(((vert[i][j]-aabb_o)/aabb_d).ptr());\
glVertex3fv(vert[i][j].ptr());

void GfxTerrainTile::pick()
{
  glLoadName(my_id << GfxObject::SUB_ID_LOG);

  if (has_cliffs) {

    // non cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    CNVp(0,0);
    CNVp(0,2);
    CNVp(0,1);
    CNVp(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(1,0);
    CNVp(1,2);
    CNVp(1,1);
    CNVp(1,3);
    glEnd(); 

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(2,0);
    CNVp(2,2);
    CNVp(2,1);
    CNVp(2,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(3,0);
    CNVp(3,2);
    CNVp(3,1);
    CNVp(3,3);
    glEnd();

    // cliff pieces
    
    glBegin(GL_TRIANGLE_STRIP);
    CNVp(0,1);
    CNVp(0,3);
    CNVp(1,0);
    CNVp(1,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(2,1);
    CNVp(2,3);
    CNVp(3,0);
    CNVp(3,2);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(2,0);
    CNVp(2,1);
    CNVp(0,2);
    CNVp(0,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(3,0);
    CNVp(3,1);
    CNVp(1,2);
    CNVp(1,3);
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
    CNVp(1,2);
    CNVp(0,3);
    CNVp(3,0);
    CNVp(2,1);
    glEnd();
  } else {

    // no cliffs

    glBegin(GL_TRIANGLE_FAN);
    CNVp(0,3);
    CNVp(0,0);
    CNVp(0,2);
    CNVp(2,0);
    CNVp(2,2);
    CNVp(2,3);
    CNVp(3,2);
    CNVp(3,3);
    CNVp(3,1);
    CNVp(1,3);
    CNVp(1,1);
    CNVp(1,0);
    CNVp(0,1);
    CNVp(0,0);
    glEnd();
  }
}

//-------------------------------------------------------------------

/** Draws an outline along the South and West sides. */

void GfxTerrainTile::draw_outline()
{
  glLoadName(my_id << GfxObject::SUB_ID_LOG);

  real4 oset = 0.01;
  
  glColor3f(0.1, 0.1, 0.1);
  glBegin(GL_LINE_STRIP);
  glVertex3f(vert[0][0].x, vert[0][0].y+oset, vert[0][0].z);
  glVertex3f(vert[0][2].x, vert[0][2].y+oset, vert[0][2].z);
  glVertex3f(vert[2][0].x, vert[2][0].y+oset, vert[2][0].z);
  glVertex3f(vert[2][2].x, vert[2][2].y+oset, vert[2][2].z);
  glVertex3f(vert[2][3].x, vert[2][3].y+oset, vert[2][3].z);
  glVertex3f(vert[3][2].x, vert[3][2].y+oset, vert[3][2].z);
  glVertex3f(vert[3][3].x, vert[3][3].y+oset, vert[3][3].z);
  glEnd();
}

//===================================================================
