/*
-----------------------------------------------------------------------------
This source file is part of OGRE
    (Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/
*/
/*
-----------------------------------------------------------------------------
Filename:    OrgeGfxModuleFrameListener_H
Description: Defines an example frame listener which responds to frame events.
This frame listener just moves a specified camera around based on
keyboard and mouse movements.
Mouse:    Freelook
W or Up:  Forward
S or Down:Backward
A:        Step left
D:        Step right
             PgUp:     Move upwards
             PgDown:   Move downwards
             F:        Toggle frame rate stats on/off
			 R:        Render mode
             T:        Cycle texture filtering
                       Bilinear, Trilinear, Anisotropic(8)
             P:        Toggle on/off display of camera position / orientation
-----------------------------------------------------------------------------
*/

#ifndef OgreGfxModuleFrameListener_H
#define OgreGfxModuleFrameListener_H

#include "Options.H"
#include "ShutDown.H"
#include "Game.H"
#include "GameChanges.H"
#include "LoopFunctor.H"
#include "SDLinit.H"

#include "TurretAI.H"
#include "GatherAI.H"
#include "KeyMap.H"

#include "ClientState.H"
#include "TerrainModule.H"

#include "GfxModule.H"
#include "SoundMixer.H"

#include <iostream>
#include "BasicObject.H"
#include "OgreGfxModuleTerrainTileGrid.H"
#include "Ogre.h"
#include "OgreKeyEvent.h"
#include "OgreEventListeners.h"
#include "OgreStringConverter.h"
#include "OgreException.h"

#include <OgreInputEvent.h>
#include <OgreNoMemoryMacros.h> 
#include <CEGUI/CEGUISystem.h>
#include <CEGUI/CEGUISchemeManager.h>
#include "OgreCEGUIRenderer.h"
#include "ObjectManager.H"

 
class OgreGfxModuleFrameListener: public Ogre::FrameListener, public Ogre::KeyListener, public Ogre::MouseListener, public Ogre::MouseMotionListener
{
private:
    int mSceneDetailIndex ;
    Ogre::Real mMoveSpeed;
    Ogre::Degree mRotateSpeed;
    Ogre::Overlay* mDebugOverlay;

  void updateStats(void);

public:
    // Constructor takes a RenderWindow because it uses that to determine input context

  /**
     Creates a new action listener. Unlike OgreGfxModule, this class should be safe to create multiple instances of.
     \param win a reference to our renderwindow that we want to handle events for
     \param cam a reference to our secene's main camera
     \param sceneManager a valid reference to our scene manager
     \param useBufferedInputKeys exactly what it sounds like
     \param useBufferedInputMouse also fairly intuitive
   */
  OgreGfxModuleFrameListener(Ogre::RenderWindow* win, Ogre::Camera* cam, Ogre::SceneManager* sceneManager, bool useBufferedInputKeys = true, bool useBufferedInputMouse = true);

  /**
     Obtains a reference to the gsm
     \param gsm the gsm
   */
  bool init(GameStateModule* gsm);

  /**
     Get rid of all allocated memory.
   */
  ~OgreGfxModuleFrameListener();
  bool processUnbufferedKeyInput(const Ogre::FrameEvent& evt);
  bool processUnbufferedMouseInput(const Ogre::FrameEvent& evt);
  void showDebugOverlay(bool show);

  /**
     call-back method called just before every frame was rendered.
     \return if flase, it's time to quit
   */
  bool frameStarted(const Ogre::FrameEvent& evt);
  bool frameEnded(const Ogre::FrameEvent& evt);
  

  void processKeyChanges();
  void switchMouseMode();
  void switchKeyMode();
  void keyClicked(Ogre::KeyEvent* e);
  void keyPressed(Ogre::KeyEvent* e);
  void keyReleased(Ogre::KeyEvent* e);
  bool processLeftMouseButton();
  bool processRightMouseButton();

  //Mouse Events.
  void mouseClicked(Ogre::MouseEvent* e);
  void mouseEntered(Ogre::MouseEvent* e);
  void mouseExited(Ogre::MouseEvent* e);
  void mousePressed(Ogre::MouseEvent* e);  // mousePressed
  void mouseReleased(Ogre::MouseEvent* e); // mouseReleased
  void mouseMoved (Ogre::MouseEvent *e);   // mouseMoved
  void mouseDragged (Ogre::MouseEvent *e); // mouseDragged

  /**
     Where the ogre gfx module
   */
  void process_changes();

  OgreGfxModuleTerrainTileGrid* battleField;

private:
  bool mRightMousePressed; 
  bool mLeftMousePressed;
  bool mRightMousePressedLastFrame;
  bool mLeftMousePressedLastFrame;
  bool mMiddleMousePressed;
  int lastMousePosX;
  int lastMousePosY;
  bool keys[256];

  ScriptObj *gui_obj;

  Ogre::RaySceneQuery* mRaySceneQuery;     // The ray scene query pointer
  Ogre::SceneManager* mSceneMgr;
  Ogre::EventProcessor* mEventProcessor;
  Ogre::InputReader* mInputDevice;
  Ogre::Camera* mCamera;

  CEGUI::System *mGUISystem; 
  CEGUI::Renderer *mGUIRenderer;     // cegui renderer
  
  Ogre::Vector3 mTranslateVector;
  Ogre::RenderWindow* mWindow;
  bool mStatsOn;
  bool mUseBufferedInputKeys, mUseBufferedInputMouse, mInputTypeSwitchingOn;
  unsigned int mNumScreenShots;
  float mMoveScale;
  Ogre::Degree mRotScale;
  // just to stop toggles flipping too fast
  Ogre::Real mTimeUntilNextToggle ;
  Ogre::Radian mRotX, mRotY;
  Ogre::TextureFilterOptions mFiltering;
  int mAniso;
  
  //orts stuff
  GameStateModule* gsm;
};

#endif
