#ifndef OgreGfxModuleTerrainTileGrid_H
#define OgreGfxModuleTerrainTileGrid_H

#define OGRE_RELEASE_ASSERT 0

#include "Ogre.h"
#include <malloc.h>
#include "OgreKeyEvent.h"
#include "OgreEventListeners.h"
#include "OgreStringConverter.h"
#include "OgreException.h"
#include <OgreStaticGeometry.h>
#include <vector>
#include "OgreGfxModuleMegaTile.H"

class OgreGfxModuleTerrainTileGrid
{
public:

  /** Creates a completely empty grid of ogre tiles*/
  OgreGfxModuleTerrainTileGrid();
  /** Frees all allocated memory*/
  ~OgreGfxModuleTerrainTileGrid();

  /** Initializes the tile grid and allocates the appropriate amount of memory
      \param mSceneMgr You must provide a valid reference to your SceneManager
      \param startingNode This node indicates where the upper-left corner of the grid will start
      \param mr this indicates how many rows all "megatiles" will have. Megatiles are large static meshes composed of regular tiles, all bound together for the purposes of performance.
      \param mc how many columns all "megatiles" will have (should probably be equal to mr).
      \param numr how many rows of Megatiles there will be.
      \param numc should probably equal numr
      \param w for scaling purposes; how wide(x direction) a regular tile is
      \param h how high (z direction) a regular tile is
      
  */
  OgreGfxModuleTerrainTileGrid(Ogre::SceneManager* mSceneMgr, Ogre::SceneNode* startingNode, int mr=4, int mc=4, int numr=16, int numc=16, float w=10.0, float h=10.0);
	
  /** Tells the fog of war what to set as the visibility of a terrain tile.
      \param r what row of the grid to set
      \param c what column
      \param bVisible if set to true, then this spot should be fully lit up
      \param bRevealed if set to true then this area will be dimly lit (unless it's currently visible too).
   */
  void setVisibility(int r, int c, bool bVisible, bool bRevealed = true);

  /** Tells the fog of war to set all visiblity values to false.
  */
  void cleanVisibleGrid();
  
  /** DEPRECEATED! Don't use!*/
  Ogre::Entity* createNewBasicTile(float MID,
				   float N, float NE, float E, float SE,
				   float S, float SW, float W, float NW,
				   float UNW_1, float VNW_1, float UNE_1, float VNE_1, 
				   float USW_1, float VSW_1, float USE_1, float VSE_1,
				   float UNW_2, float VNW_2, float UNE_2, float VNE_2, 
				   float USW_2, float VSW_2, float USE_2, float VSE_2);
  
  /** Use this method to create a new Ogre terrain tile entity
      \param vertices all the vertex information (i.e., pos, normals, uvw1, and uvw2 )
      \param nVertices how many vertices are in vertices[]; currently may only be 9 or 36
      \return an Ogre::Entity* which is a reference to the newly created terrain tile mesh
  */
  Ogre::Entity* createNewTileWithTheseVerticies(float vertices[], int nVertices = 9);

  /**Use this to add an entity to the terrain. NOTE: After adding all the entities you want, you MUST
     call refreshGeometry() in order to have it alll appear in the scene!
     \param tiler tells which megatile to attach the tile to
     \param tilec which column
     \param newOgreTile a reference to the ogre entity you whish to add to the terrain.
  */
  void addTileToGeometry(int tiler, int tilec, Ogre::Entity* newOgreTile);
  
  /** The fog of war changes over time with a fade effect. This method should be called about 8 to 10 times a second to create a cool-looking fog-fading effect.
      \param fadeAmount a value between 1 and 255, 16 is about right.
  */
  void updateTemporalFade(int fadeAmount);//do a cool fog-of-war fading effect
  
  /**You shouldn't need to ever call this. Instead use updateTemporalFade(...)*/
  void refreshFogOfWar();//make sure that the fog of war is currently accurate
 
  /** After adding new geometry, you mush call this to rebuild the geometry so that the new stuff will appear.*/
  void refreshGeometry();
  
  int rows;//The TOTAL number of rows (regardless of the megaTile properties)
  int columns;//The TOTAL number of columns (regardless of the megaTile properties)
  int nRows;
  int nCols;
  int mRows;
  int mCols;
  float tileWidth;
  float tileHeight;
  Ogre::String tileName;
  bool bRebuildNeeded;
  bool bFogOfWarRefreshNeeded;
  int tileCount;
  bool** revealed;
  bool** visible;

private:
  Ogre::SceneNode* startNode;
  //Ogre::StaticGeometry* battlefieldMap;
  Ogre::SceneManager* mSceneMgr;
  int** alphaVals;
  //int** mapVals;
  void prepareFogOfWarTextureCoordinates(float vertices[], int nVertices);
  OgreGfxModuleMegaTile** megaTiles;
  int fogOfWarTexturePixelsWidth;
  int fogOfWarTexturePixelsHeight;
  Ogre::String entityName;
  Ogre::MaterialPtr terrainMaterial;
  Ogre::String terrainMaterialName;
  Ogre::TexturePtr fogOfWartexture;
  
  void prepareFogOfWar(unsigned int widthInPixels,unsigned int heightInPixels,Ogre::String texturesName);
  void prepareTerrainMaterial(Ogre::String fogOfWarName, Ogre::String landscapeName);
  void synchronizeVisibilityGrid();
  
};


#endif
