/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "RTS_Game.H"
#include "Element.H"
#include "BoolMatrix.H"
#include "Compress.H"

#define COMPRESS     1
#define DBG_CHECKSUM 0

bool RTS_Game::check_param(ostream &oserr, RT_GameParam &param)
{
  return false;
}

static void compress_ostream(Compress &compress, ostringstream &os_msg, vector<string> &view)
{
  view.clear();

  const sint4 MSG_SIZE = 250;
  const char *msg = os_msg.str().c_str();
  sint4 msg_n = os_msg.str().size();

  if (msg_n == 0) ERR("empty ostream");
  
#if COMPRESS

  const sint4 BLOCK_LEN = 64000; // < max. string size (64K?)

  FORS (i, (msg_n+BLOCK_LEN-1)/BLOCK_LEN) {
    sint4 start = i * BLOCK_LEN;
    sint4 end   = start + BLOCK_LEN;
    if (end > msg_n) end = msg_n;

    // compress block

    string sin(msg+start, end-start);
    string sout;
    compress.doit(sin, sout);
    sint4 out_size = sout.size();
    const char *so = sout.c_str();

    // cout << "in= " << (end-start) << " out=" << out_size << endl;

    sint4 msg_number = (out_size+MSG_SIZE-1)/MSG_SIZE;
    
    FORS (i, msg_number) {
      sint4 start = i * MSG_SIZE;
      sint4 end = start + MSG_SIZE;
      if (end > out_size) end = out_size;
      string s;
      if (i == msg_number-1) s = "1"; else s = "0"; // 1: last message in block
      s.append(so+start, end-start);

      // compressed object messages start with '0' or '1'
      view.push_back(s);
    }
  }
  
#else

  sint4 mn = (msg_n+MSG_SIZE-1)/MSG_SIZE;

  FORS (i, mn) {
    sint4 start = i * MSG_SIZE;
    sint4 end   = start + MSG_SIZE;
    if (end > msg_n) end = msg_n;
    string s(msg+start, end-start);
    view.push_back(s);
  }
  
#endif

  // os_msg.rdbuf()->freeze(false);
}


static void decompress_strings(Decompress &decompress,
			       const vector<string> &strs, ostream &os)
{
  uint4 no = strs.size();

  if (no == 0) ERR("empty string vector");
  
#if COMPRESS
  
  string block;
  
  FORU (i, no) {
    assert(strs.size() > 0);
    // skip 0/1 prefix
    block.append(strs[i].c_str()+1, strs[i].size()-1);
    if (strs[i][0] == '1') {
      // block complete -> decompress
      string sout;
      if (decompress.doit(block, sout)) ERR("decompress error");
      os << sout;
      block.clear();
    } else if (strs[i][0] != '0') ERR("0/1 expected");
  }

#else
  
  FORU (i, no) os << strs[i];

#endif

}


// make sure that object messages don't start with '#'
// call update_prev_prop() after creating all views!

void RTS_Game::create_view_diff(Compress &compress, sint4 player, 
				vector<string> &view, uint4 &checksum)
{
  ostringstream os_msg;
  uint4 n = elems.size();
  set<uint4> visible_ind;
  sint4 pnum;

  opt.get("-pnum", pnum);

  assert(player >= 0);

  checksum = 0;

#if DBG_CHECKSUM  
  FORU (i, n) {
    cout << "prev " << i << " ";
    elems[i]->prev_prop.write();
    cout << endl;
  }
#endif
  
  bool shared[pnum];
  FORS (i, pnum) shared[i] = x_shares_view_with_y->get(i, player);
  shared[player] = true;

  RTS_PlayerInfo &pinf = players[player];
  
  FORU (i, n) {
    if (shared[elems[i]->sv_owner]) {
      visible_ind.insert(i);
      NeighborInfo &ni = elems[i]->neighbors;
      sint4 nn = ni.size();
      FORS (j, nn) {
	if (elems[ni[j]]->sv_owner != player) visible_ind.insert(ni[j]);
      }
    }
  }

  set<ObjId> curr_obj_cids;
  IdMap id2ind;
 
  FORALL (set<uint4>, visible_ind, it) {
    ObjId sid = elems[*it]->sv_id;
    ObjId cid = pinf.get_cid(sid);
    if (!cid) cid = pinf.add_sid(sid);
    // cout << "add cid " << cid << " sid=" << sid << endl;
    curr_obj_cids.insert(cid);
    id2ind[cid] = *it;
  }

#ifndef NDEBUG  
  pinf.check();
#endif
  
  view.clear();

  String DEL_STR, NEW_STR, UPD_STR;
  UPD_STR.form("%c", UPD_CHAR);
  DEL_STR.form("%c", DEL_CHAR);
  NEW_STR.form("%c", NEW_CHAR);
  
  set<ObjId>::iterator old_it  = pinf.prev_obj_cids.begin();
  set<ObjId>::iterator old_end = pinf.prev_obj_cids.end();

  os_msg << " "; // create at least one character
  
  FORALL (set<ObjId>, curr_obj_cids, i) {

    // cout << "CURR_ID " << (*i) << endl;
    
    for (; old_it != old_end && *old_it < *i; old_it++) {
      pinf.delete_cid(*old_it);
      os_msg << DEL_STR << " ";
    }

    bool new_obj = false;
    
    if (old_it != old_end && *old_it == *i) {
      os_msg << UPD_STR << " "; old_it++;
    } else {
      os_msg << NEW_STR << " " << (*i) << " ";
      new_obj = true;
    }
    IdMap::const_iterator mit = id2ind.find(*i);
    if (mit == id2ind.end()) ERR("index not found");

#if DEBUG_CHECKSUM    
    cout << "INDEX=" << (*mit).second << endl;
#endif
    
    Element &e = *elems[(*mit).second];
    Prop prev_view, curr_view;

    // compute previous view

    if (e.prev_seen_by.find(player) != e.prev_seen_by.end()) {

      assert(!new_obj);

#if DEBUG_CHECKSUM      
      cout << "seen before" << endl;
#endif
      
      // e was seen by player before

      prev_view = e.prev_prop;
      if (!prev_x_shares_view_with_y->get(e.sv_owner, player)) {
	prev_view.opp_view();
      }

    } else {
 
#if DEBUG_CHECKSUM      
      cout << "not seen before" << endl;
#endif
      
      // else default view
    }

    curr_view = (Prop)e;
    assert(e.seen_by.find(player) != e.seen_by.end());
    if (!x_shares_view_with_y->get(e.sv_owner, player)) curr_view.opp_view();

#if DBG_CHECKSUM
    cout << "prev view:" << player << " "; prev_view.write(); cout << endl;    
    cout << "curr view:" << player << " "; curr_view.write(); cout << endl;
#endif
    
    checksum += curr_view.checksum(*i); // same as in elems_checksum!

    curr_view.gen_diff(prev_view, os_msg);
    os_msg << " ";
  }

#if DBG_CHECKSUM  
  cout << "CS=" << checksum << endl;
#endif
  
  for (; old_it != old_end; old_it++) {
    pinf.delete_cid(*old_it);
    os_msg << DEL_STR << " ";
  }

  compress_ostream(compress, os_msg, view);
  pinf.prev_obj_cids = curr_obj_cids;

#if DBG_CHECKSUM
  FORU (i, n) {
    cout << "prev! " << i << " ";
    elems[i]->prev_prop.write();
    cout << endl;
  }
  FORU (i, n) {
    cout << "prev$ " << i << " ";
    elems[i]->Prop::write();
    cout << endl;
  }
#endif  
}


void RTS_Game::update_prev_prop()
{
  uint4 n = elems.size();
  FORU (i, n) elems[i]->prev_prop = *elems[i];
}


void RTS_Game::clear_actions()
{
  uint4 n = elems.size();
  Action no_act;
  
  FORU (i, n) elems[i]->last_action = no_act;
}


void RTS_Game::update_view(Decompress &decompress, const vector<string> &obj_msgs)
{
  uint4 n = elems.size();
  set<ObjId> curr_cind;
  ostringstream os_msg;

  decompress_strings(decompress, obj_msgs, os_msg);
  istringstream is(os_msg.str());

  FORU (i, n) {
    Element &e = *elems[i];
    e.prev_prop = static_cast<Prop>(e);
    ObjId cid = e.sv_id;
    curr_cind.insert(cid);
  }

  set<ObjId>::iterator curr_it  = curr_cind.begin();
  set<ObjId>::iterator curr_end = curr_cind.end();
  IdMap::const_iterator mit;

  //cout << n << " ELEMENT(S) before" << endl;
  
  FOREVER {

    char c;

    is >> c;
    if (is.eof()) break;
    if (!is) ERR("corrupt msg A");

    switch (c) {

    case UPD_CHAR: // update element with index *curr_it

      {
	if (curr_it == curr_end) ERR("FOO");
	mit = id2index.find(*curr_it);
	if (mit == id2index.end()) ERR("FOO");
	elems[(*mit).second]->apply_diff(is);
	if (!is) ERR("FOO");
	curr_it++;	
      }
      break;
    
    case NEW_CHAR: // create new object

      {
	Element *pe = new Element();
	is >> pe->sv_id;
	if (!is) ERR("id expected");
	pe->apply_diff(is);
	if (!is) ERR("corrupt diff");
	elems.push_back(pe);
	mit = id2index.find(pe->sv_id);
	if (mit != id2index.end()) ERR("new: duplicate id!");

	//cout << "NEW " << pe->sv_id << endl;
      }
      break;
	
    case DEL_CHAR: // delete obj (free spot has to be filled at the end!)

      //cout << "DEL " << (*curr_it) << endl;

      if (curr_it == curr_end) ERR("del: end reached");
      mit = id2index.find(*curr_it);
      if (mit == id2index.end()) ERR("del: id not found");
      delete elems[(*mit).second];
      elems[(*mit).second] = 0;
      curr_it++;
      break;
    
    default: ERR("corrupt msg B");
    }
  }

  if (curr_it != curr_end) ERR("not all ids visited");

  //cout << "-----------" << endl;
  
  // fill deleted spots
  
  uint4 en = elems.size();

  FORU (i, en) {
    if (!elems[i]) {
      elems[i] = elems[en-1];
      elems.pop_back();      
      en--; i--; 
    }
  }
  
  // os_msg.rdbuf()->freeze(false);

  refresh_id2index();
}

  
void RTS_Game::refresh_id2index()
{
  id2index.clear();
  uint4 n = elems.size();
  FORU (i, n) id2index[elems[i]->sv_id] = i;
  if (id2index.size() != n) ERR("duplicate ids");
}

void RTS_Game::prop_vectors(vector<Prop> &objs) const
{
  uint4 n = elems.size();

  objs.clear();
  objs.reserve(n);

  FORU (i, n) objs.push_back(*elems[i]);
}


uint4 RTS_Game::elems_checksum()
{
  uint4 n = elems.size();
  uint4 cs = 0;

  FORU (i, n) {
    Prop *pp = static_cast<Prop*>(elems[i]);
    cs += pp->checksum(elems[i]->sv_id);
#if DBG_CHECKSUM
    pp->write(); cout << endl;
#endif    
  }

#if DBG_CHECKSUM  
  cout << "CS=" << cs << endl;
#endif  
  return cs;
}


struct IdPair { ObjId id; uint4 index; };
  
struct IdPairComp
{
  bool operator()(const IdPair &p1, const IdPair &p2) const
  {
    if (p1.id < p2.id) return true;
    if (p2.id < p1.id) return false;
    if (&p1 == &p2) return false;
    ERR("ids not unique!");
    return false;
  }
  
};


bool RTS_Game::set_action(sint4 player, uint4 id, Action &act)
{
  map<uint4,uint4>::const_iterator mit = id2index.find(id);
  if (mit == id2index.end()) {
    ERR("id not found");
    return true;
  }
  
  Element &e = *elems[(*mit).second];
  if (e.sv_owner != player) {
    ERR("not ownver");
    return true;
  }
  e.last_action = act;
  return false;
}


void RTS_Game::create_action_messages(Compress &compress, sint4 cpind, vector<string> &msgs)
{
  uint4 n = elems.size();
  set<IdPair, IdPairComp> pair_set;

  FORU (i, n) {
    if (elems[i]->sv_owner == cpind) {
      IdPair pair;
      pair.id = elems[i]->sv_id;
      pair.index = i;
      // cout << "(" << i << "," << pair.id << ")" << endl;
      pair_set.insert(pair);
    }
  }

  ostringstream os_msg;

  os_msg << " "; // create at least one character
  
  FORALL (set<IdPair>, pair_set, it) {
    elems[(*it).index]->last_action.write(os_msg);
    os_msg << " ";
  }

  msgs.clear();
  compress_ostream(compress, os_msg, msgs);
}


void RTS_Game::update_actions(Decompress &decompress, sint4 cpind, vector<string> &act_msgs)
{
  uint4 n = elems.size();
  set<IdPair, IdPairComp> pair_set;
  
  FORU (i, n) {
 
    if (elems[i]->sv_owner == cpind) {
      IdPair pair;
      pair.id = elems[i]->sv_id;
      pair.index = i;
      pair_set.insert(pair);
    }
  }
  
  ostringstream os_msg;
  decompress_strings(decompress, act_msgs, os_msg);
  istringstream is(os_msg.str());
  map<uint4, uint4> &cid2sid = players[cpind].cid2sid;

  FORALL (set<IdPair>, pair_set, it) {
    Action &act = elems[(*it).index]->last_action;
    act.read(is);
    if (!is) ERR("action expected");
    if (act.action == Action::ATTACK) {

      // map client target_id to server target_id
      
      map<uint4, uint4>::const_iterator mit = cid2sid.find(act.target_id);
      if (mit == cid2sid.end()) {
	cerr << "invalid target!" << endl;
	act.action = Action::NO_ACTION;
      } else {
	act.target_id = (*mit).second;
      }
    }
  }

  uint4 no;
  is >> no;
  if (!is.eof()) ERR("too many messages");
  //os_msg.rdbuf()->freeze(false);
}


void RTS_Game::create_vicinities(sint4 cpind, vector< vector<Prop> > &vics)
{
  vics.clear();
  uint4 n = elems.size();
  
  compute_visible_elements();
  
  FORU (i, n) {
    if (elems[i]->sv_owner == cpind) {
      vector<Prop> props;

      props.push_back(*elems[i]);  // obj. itself
      NeighborInfo &ni = elems[i]->neighbors;
      sint4 nn = ni.size();
      FORS (j, nn) props.push_back(*elems[ni[j]]);
      vics.push_back(props);
    }
  }
}

