/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "SocketBuffer.H"
#include "System.H"

#define TEST 0

#define ALARM_ON(i)
#define ALARM_OFF

// return < 0 <=> error
// == 0 : nothing or timeout
// > 0  :  data read

sint4 SocketBuffer::read_socket(bool blocking)
{
  fd_set         fds;
  sint4          i;
  struct timeval timeout, *ptime;

  if (blocking) ptime = 0;
  else {
    timeout.tv_sec  = 0;
    timeout.tv_usec = 0;
    ptime = &timeout;
  }
  
  FD_ZERO(&fds);
  FD_SET(my_socket, &fds);
  ALARM_ON(u)
  i = System::select_desc(my_socket+1, fds, *(fd_set*)0, *(fd_set*)0,
			  ptime, false);

  // printf("read select: %d\n", i);

  ALARM_OFF

  if (i <= 0) {

    if (i < 0) { cout << i << endl; perror("FOO1"); }
    return i; // error or no progress
  }

  // data ready to read

  // move data to front (if not already there) to make room

  if (ipos != 0) {
    memmove(ibuff, ibuff+ipos, ilen);
    ipos = 0;
  }
  
  sint4 rem_len = ibufflen-ilen;
  
  if (rem_len <= 0) return 0; // buffer full (should have been read!)
  
  ALARM_ON(u)
  sint4 n = System::sock_read(my_socket, ibuff+ilen, rem_len, false);
  ALARM_OFF
  if (n <= 0) return -1;

  ilen += n;
  return n;  // progress
}


sint4 SocketBuffer::read(char *msg, sint4 len, bool blocking)
{
  if (len <= 0) ERR("len <= 0");
  if (len > ibufflen) ERR("len > IBUFFLEN");

  while (ilen < len) {
    sint4 res = read_socket(blocking); // try reading data
    if (res < 0) return res;
    if (!blocking) break;
  }
  
  if (ilen < len) return 0;  // no input

  memcpy(msg, ibuff+ipos, len);
  ipos += len;
  ilen -= len;
  return len;
}


sint4 SocketBuffer::write_socket(bool blocking)
{
  sint4 used = o_used();
  
  if (used == 0) return 0; // nothing to do

  sint4 n;
  sint4 begin, len;
  
  if (ostart < oend) {

    assert(oend-ostart == used);

    len = used;
    
  } else {

    assert(o_used() == obufflen-ostart+oend);
    assert(ostart != 0);

    begin = ostart;
    len = obufflen - ostart;
  }

  n = System::sock_write(my_socket, obuff+ostart, len, false);

  if (n < 0) return n;  // error

  if (n == 0 && blocking) {

    fd_set fds;
    FD_ZERO(&fds);
    FD_SET(my_socket, &fds);
    ALARM_ON(u)
    int ret = System::select_desc(my_socket+1, *(fd_set*)0, fds, *(fd_set*)0,
				  0, false);
    ALARM_OFF
      
    if (ret <= 0) return ret; // select error or blocked

    // try again

    n = System::sock_write(my_socket, obuff+ostart, len, false);
  }

  if (n <= 0) return n;

  ostart += n;
  if (ostart >= obufflen) ostart = 0;

  return n;
}


sint4 SocketBuffer::write(const char *msg, sint4 len, bool blocking)
{
  if (len <= 0) ERR("len <= 0");
  if (len > obufflen-1) ERR("len > OBUFFLEN");

  while (len > o_available()) {

    if (len > o_available()) {
      sint4 res = write_socket(blocking); // try writing data
      if (res < 0) return res;
    }

    if (!blocking) break;
  }
  
  sint4 a = o_available();

  if (len > a) return 0;   // output blocked

  if (oend+len <= obufflen) {

    memcpy(obuff+oend, msg, len);
    oend += len;
    if (oend >= obufflen) {
      oend -= obufflen;
      assert(oend == 0);
    }

  } else {

    memcpy(obuff+oend, msg, obufflen-oend);
    memcpy(obuff, msg+obufflen-oend, len-(obufflen-oend));
    oend += len - obufflen;

    assert(oend >= 0 && oend < obufflen);
  }

  assert(o_available() == a - len);
  
  return len;
}


sint4 SocketBuffer::blocking_flush() {

  while (o_used() > 0) {

    // write remaining data
    
    sint4 res = write_socket(true);
    if (res < 0) return res;
  }

  return 1;
}

#if TEST
static void write_binstr(const string &msg, ostream &os=cout) {
  FORU (i, msg.size()) {
    uint4 c = (msg[i] & 0xff);
    if (c < 0x20 || c >= 0x7f) form(os, "<%x>", c);
    else os << (char)c;
  }
}
#endif

sint4 SocketBuffer::write_msg(const string &msg, bool blocking)
{
#if TEST
  cout << "write_msg: <"; write_binstr(msg); cout << ">" << endl;
#endif
  
  uint4 l = msg.size();
  assert(l < 255);
  char s[l+1];
  memcpy(s+1, msg.c_str(), l);
  s[0] = l;
  return write(s, l+1, blocking);
}


sint4 SocketBuffer::read_msg(string &msg, bool blocking)
{
  sint4 ret;

  while (msg_len == NO_MSG_LEN) {

    ret = read(msg_len, blocking);
    if (ret <= 0) return ret;
    if (msg_len == NO_MSG_LEN || msg_len == 0) ERR("msg-len corrupt");
    if (!blocking) break;
  }

  char s[msg_len+1];
  ret = read(s, msg_len, blocking);
  if (ret <= 0) return ret;

  msg.clear();
  msg.append(s, msg_len);
  msg_len = NO_MSG_LEN;

#if TEST
  cout << "read_msg: <"; write_binstr(msg); cout << ">" << endl;
#endif
  return 1;
}

sint4 SocketBuffer::read_socket2(bool ready)
{
  // move data to front (if not already there) to make room

  if (ipos != 0) {
    memmove(ibuff, ibuff+ipos, ilen);
    ipos = 0;
  }
  
  sint4 rem_len = ibufflen-ilen;
  
  if (rem_len <= 0) return 0; // buffer full (should have been read!)

  sint4 n = System::sock_read(my_socket, ibuff+ilen, rem_len, false);

  if (ready && n == 0) return -1;
  if (n <= 0) return n;

  ilen += n;
  return n;  // progress
}


sint4 SocketBuffer::read2(char *msg, sint4 len, bool ready)
{
  if (len <= 0) ERR("len <= 0");
  if (len > ibufflen) ERR("len > IBUFFLEN");

  if (ilen < len) {
    sint4 res = read_socket2(ready);   // try reading data
    if (res < 0) return res;           // error or connection closed
  }
  
  if (ilen < len) return 0;  // not enough

  memcpy(msg, ibuff+ipos, len);
  ipos += len;
  ilen -= len;
  return len;
}


sint4 SocketBuffer::read_msg2(string &msg, bool ready)
{
  sint4 ret;

  if (msg_len == NO_MSG_LEN) {
    ret = read(msg_len, ready);
    if (ret <= 0) return ret;
    if (msg_len == NO_MSG_LEN || msg_len == 0) return -1;
    ready = false;
  }

  char s[msg_len+1];
  ret = read2(s, msg_len, ready);
  if (ret <= 0) return ret;

  msg.clear();
  msg.append(s, msg_len);
  msg_len = NO_MSG_LEN;

#if TEST
  cout << "read_msg: <"; write_binstr(msg); cout << ">" << endl;
#endif
  return 1;
}
